"""
Dropbox OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dropbox.html
"""

from .oauth import BaseOAuth2


class DropboxOAuth2V2(BaseOAuth2):
    name = "dropbox-oauth2"
    ID_KEY = "uid"
    AUTHORIZATION_URL = "https://www.dropbox.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://api.dropboxapi.com/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Dropbox account"""
        name = response.get("name")
        return {
            "username": str(response.get("account_id")),
            "email": response.get("email"),
            "fullname": name.get("display_name"),
            "first_name": name.get("given_name"),
            "last_name": name.get("surname"),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.dropboxapi.com/2/users/get_current_account",
            headers={"Authorization": f"Bearer {access_token}"},
            method="POST",
        )



# pyright: reportAttributeAccessIssue=false

import json

from httpretty import HTTPretty

from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .test_open_id_connect import OpenIdConnectTestMixin

ROOT_URL = "https://cas.example.net/"


class CASOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.cas.CASOpenIdConnectAuth"
    issuer = f"{ROOT_URL}oidc"
    openid_config_body = json.dumps(
        {
            "issuer": f"{ROOT_URL}oidc",
            "jwks_uri": f"{ROOT_URL}oidc/jwks",
            "authorization_endpoint": f"{ROOT_URL}oidc/oidcAuthorize",
            "token_endpoint": f"{ROOT_URL}oidc/oidcAccessToken",
            "userinfo_endpoint": f"{ROOT_URL}oidc/oidcProfile",
            "request_uri_parameter_supported": False,
            "grant_types_supported": ["authorization_code"],
            "token_endpoint_auth_methods_supported": ["client_secret_basic"],
        }
    )

    expected_username = "cartman"

    user_data_body = json.dumps(
        {
            "sub": "Cartman",
            "service": "https://cas.example.net/complete/cas/",
            "auth_time": 1677057708,
            "attributes": {
                "name": "Eric",
                "groups": ["users", "admins"],
                "preferred_username": "cartman",
                "email": "cartman@example.net",
            },
            "id": "Cartman",
            "client_id": "dev",
        }
    )

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {
                f"SOCIAL_AUTH_{self.name}_OIDC_ENDPOINT": f"{ROOT_URL}oidc",
            }
        )
        return settings

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        HTTPretty.register_uri(
            "GET",
            uri=self.backend.userinfo_url(),
            status=200,
            body=self.user_data_body,
            content_type="text/json",
        )

    def test_everything_works(self):
        self.do_login()



"""
Kakao OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/kakao.html
"""

from .oauth import BaseOAuth2


class KakaoOAuth2(BaseOAuth2):
    """Kakao OAuth authentication backend"""

    name = "kakao"
    AUTHORIZATION_URL = "https://kauth.kakao.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://kauth.kakao.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("properties", "properties"),
    ]

    def get_user_id(self, details, response):
        return response["id"]

    def get_user_details(self, response):
        """Return user details from Kakao account"""

        kaccount_email = ""
        kakao_account = response.get("kakao_account", "")
        if kakao_account:
            kaccount_email = kakao_account.get("email", "")
        properties = response.get("properties", "")
        nickname = properties.get("nickname") if properties else ""
        return {
            "username": nickname,
            "email": kaccount_email,
            "fullname": nickname,
            "first_name": nickname[1:] if nickname else "",
            "last_name": nickname[0] if nickname else "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://kapi.kakao.com/v2/user/me",
            headers={
                "Authorization": f"Bearer {access_token}",
                "Content_Type": "application/x-www-form-urlencoded;charset=utf-8",
            },
            params={"access_token": access_token},
        )

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",
            "code": self.data.get("code", ""),
            "client_id": client_id,
            "client_secret": client_secret,
        }



"""
MediaWiki OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mediawiki.html
"""

import re
import time
from urllib.parse import parse_qs, urlencode, urlparse

import jwt
import requests
from requests_oauthlib import OAuth1

from ..exceptions import AuthException
from .oauth import BaseOAuth1


def force_unicode(value):
    """
    Return string in unicode.
    """
    if isinstance(value, str):
        return value
    return str(value, "unicode-escape")


class MediaWiki(BaseOAuth1):
    """
    Handles the handshake with Mediawiki and fetching of user data.
    """

    name = "mediawiki"
    MEDIAWIKI_URL = "https://meta.wikimedia.org/w/index.php"
    SOCIAL_AUTH_MEDIAWIKI_CALLBACK = "oob"
    LEEWAY = 10.0

    def unauthorized_token(self):
        """
        Return request for unauthorized token (first stage)

        Mediawiki request token is requested from e.g.:
         * https://en.wikipedia.org/w/index.php?title=Special:OAuth/initiate
        """
        params = self.request_token_extra_arguments()
        params.update(self.get_scope_argument())
        params["title"] = "Special:OAuth/initiate"
        key, secret = self.get_key_and_secret()
        response = self.request(
            self.setting("MEDIAWIKI_URL"),
            params=params,
            auth=OAuth1(key, secret, callback_uri=self.setting("CALLBACK")),
            method=self.REQUEST_TOKEN_METHOD,
        )

        if response.content.decode().startswith("Error"):
            raise AuthException(self, response.content.decode())

        return response.content.decode()

    def oauth_authorization_request(self, token):
        """
        Generates the URL for the authorization link
        """
        if not isinstance(token, dict):
            token = parse_qs(token)
        oauth_token = token.get(self.OAUTH_TOKEN_PARAMETER_NAME)[0]  # type: ignore[reportOptionalSubscript]
        state = self.get_or_create_state()
        base_url = self.setting("MEDIAWIKI_URL")

        return "{}?{}".format(
            base_url,
            urlencode(
                {
                    "title": "Special:Oauth/authenticate",
                    self.OAUTH_TOKEN_PARAMETER_NAME: oauth_token,
                    self.REDIRECT_URI_PARAMETER_NAME: self.get_redirect_uri(state),
                }
            ),
        )

    def access_token(self, token):
        """
        Fetches the Mediawiki access token.
        """
        auth_token = self.oauth_auth(token)

        response = requests.post(
            url=self.setting("MEDIAWIKI_URL"),
            params={"title": "Special:Oauth/token"},
            auth=auth_token,
        )
        if response.content.decode().startswith("Error"):
            raise AuthException(self, response.content.decode())
        credentials = parse_qs(response.content)
        oauth_token_key = credentials.get(b"oauth_token")[0]  # type: ignore[reportOptionalSubscript]
        oauth_token_secret = credentials.get(b"oauth_token_secret")[0]  # type: ignore[reportOptionalSubscript]
        oauth_token_key = oauth_token_key.decode()
        oauth_token_secret = oauth_token_secret.decode()

        return {
            "oauth_token": oauth_token_key,
            "oauth_token_secret": oauth_token_secret,
        }

    def get_user_details(self, response):
        """
        Gets the user details from Special:OAuth/identify
        """
        key, secret = self.get_key_and_secret()
        access_token = response["access_token"]

        auth = OAuth1(
            key,
            client_secret=secret,
            resource_owner_key=access_token["oauth_token"],
            resource_owner_secret=access_token["oauth_token_secret"],
        )

        req_resp = requests.post(
            url=self.setting("MEDIAWIKI_URL"),
            params={"title": "Special:OAuth/identify"},
            auth=auth,
        )

        try:
            identity = jwt.decode(
                req_resp.content,
                secret,
                audience=key,
                algorithms=["HS256"],
                leeway=self.LEEWAY,
            )
        except jwt.InvalidTokenError as exception:
            raise AuthException(
                self,
                "An error occurred while trying to read json "
                + f"content: {exception}",
            )

        issuer = urlparse(identity["iss"]).netloc
        expected_domain = urlparse(self.setting("MEDIAWIKI_URL")).netloc

        if not issuer == expected_domain:
            raise AuthException(
                self,
                f"Unexpected issuer {issuer}, expected {expected_domain}",
            )

        now = time.time()
        issued_at = float(identity["iat"])
        if not now >= (issued_at - self.LEEWAY):
            raise AuthException(
                self, f"Identity issued {issued_at - now} seconds in the future"
            )

        authorization_header = force_unicode(req_resp.request.headers["Authorization"])
        request_nonce = re.search(r'oauth_nonce="(.*?)"', authorization_header).group(1)

        if identity["nonce"] != request_nonce:
            raise AuthException(
                self,
                "Replay attack detected: {} != {}".format(
                    identity["nonce"], request_nonce
                ),
            )

        return {
            "username": identity["username"],
            "userID": identity["sub"],
            "email": identity.get("email"),
            "confirmed_email": identity.get("confirmed_email"),
            "editcount": identity.get("editcount"),
            "rights": identity.get("rights"),
            "groups": identity.get("groups"),
            "registered": identity.get("registered"),
            "blocked": identity.get("blocked"),
        }

    def get_user_id(self, details, response):
        """
        Get the unique Mediawiki user ID.
        """
        return details["userID"]



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class KhanAcademyOAuth1Test(OAuth1Test):
    backend_path = "social_core.backends.khanacademy.KhanAcademyOAuth1"
    user_data_url = "https://www.khanacademy.org/api/v1/user"
    expected_username = "foo@bar.com"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "email": "foo@bar.com",
            "user_id": "http://googleid.khanacademy.org/11111111111111",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
OpenInfra OpenId backend
"""

from urllib.parse import urlsplit

from openid.extensions import ax

from .open_id import OpenIdAuth


class OpenInfraOpenId(OpenIdAuth):
    name = "openinfra"
    URL = "id.openinfra.dev"

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values["username"] = values.get("username") or urlsplit(
            response.identity_url
        ).path.strip("/")
        values["nickname"] = values.get("nickname", "")
        return values

    def setup_request(self, params=None):
        """Fetch email, firstname, lastname from openid"""
        request = self.openid_request(params)

        # TODO: use sreg instead ax request to fetch nickname as username
        fetch_request = ax.FetchRequest()
        fetch_request.add(
            ax.AttrInfo(
                "http://axschema.org/contact/email", alias="email", required=True
            )
        )

        fetch_request.add(
            ax.AttrInfo(
                "http://axschema.org/namePerson/first", alias="firstname", required=True
            )
        )

        fetch_request.add(
            ax.AttrInfo(
                "http://axschema.org/namePerson/last", alias="lastname", required=True
            )
        )

        request.addExtension(fetch_request)
        return request



"""
Mail.ru OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mailru.html
"""

from hashlib import md5
from urllib.parse import unquote

from .oauth import BaseOAuth2


class MailruOAuth2(BaseOAuth2):
    """Mail.ru authentication backend"""

    name = "mailru-oauth2"
    ID_KEY = "uid"
    AUTHORIZATION_URL = "https://connect.mail.ru/oauth/authorize"
    ACCESS_TOKEN_URL = "https://connect.mail.ru/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [("refresh_token", "refresh_token"), ("expires_in", "expires")]

    def get_user_details(self, response):
        """Return user details from Mail.ru request"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=unquote(response["first_name"]),
            last_name=unquote(response["last_name"]),
        )
        return {
            "username": unquote(response["nick"]),
            "email": unquote(response["email"]),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Mail.ru REST API"""
        key, secret = self.get_key_and_secret()
        data = {
            "method": "users.getInfo",
            "session_key": access_token,
            "app_id": key,
            "secure": "1",
        }
        param_list = sorted(f"{item}={value}" for item, value in data.values())
        data["sig"] = md5(("".join(param_list) + secret).encode("utf-8")).hexdigest()
        return self.get_json("http://www.appsmail.ru/platform/api", params=data)[0]


class MRGOAuth2(BaseOAuth2):
    name = "mailru"
    ID_KEY = "email"
    AUTHORIZATION_URL = "https://oauth.mail.ru/login"
    ACCESS_TOKEN_URL = "https://oauth.mail.ru/token"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [("refresh_token", "refresh_token"), ("expires_in", "expires")]
    REDIRECT_STATE = False

    def get_user_details(self, response):
        return {
            "gender": response.get("gender"),
            "fullname": response.get("name"),
            "username": response.get("name"),
            "first_name": response.get("first_name"),
            "last_name": response.get("last_name"),
            "locale": response.get("locale"),
            "email": response.get("email"),
            "address": response.get("address"),
            "birthday": response.get("birthday"),
            "image": response.get("image"),
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://oauth.mail.ru/userinfo", params={"access_token": access_token}
        )



"""
Deezer backend, docs at:
    https://developers.deezer.com/api/oauth
    https://developers.deezer.com/api/permissions
"""

from urllib.parse import parse_qsl

from .oauth import BaseOAuth2


class DeezerOAuth2(BaseOAuth2):
    """Deezer OAuth2 authentication backend"""

    name = "deezer"
    ID_KEY = "name"
    AUTHORIZATION_URL = "https://connect.deezer.com/oauth/auth.php"
    ACCESS_TOKEN_URL = "https://connect.deezer.com/oauth/access_token.php"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "app_id": client_id,
            "secret": client_secret,
            "code": self.data.get("code"),
        }

    def request_access_token(self, *args, **kwargs):
        response = self.request(*args, **kwargs)
        return dict(parse_qsl(response.text))

    def get_user_details(self, response):
        """Return user details from Deezer account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("firstname"), last_name=response.get("lastname")
        )
        return {
            "username": response.get("name"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "http://api.deezer.com/user/me", params={"access_token": access_token}
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MixcloudOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.mixcloud.MixcloudOAuth2"
    user_data_url = "https://api.mixcloud.com/me/"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "username": "foobar",
            "cloudcast_count": 0,
            "following_count": 0,
            "url": "http://www.mixcloud.com/foobar/",
            "pictures": {
                "medium": "http://images-mix.netdna-ssl.com/w/100/h/100/q/85/"
                "images/graphics/33_Profile/default_user_600x600-v4.png",
                "320wx320h": "http://images-mix.netdna-ssl.com/w/320/h/320/q/85/"
                "images/graphics/33_Profile/"
                "default_user_600x600-v4.png",
                "extra_large": "http://images-mix.netdna-ssl.com/w/600/h/600/q/85/"
                "images/graphics/33_Profile/"
                "default_user_600x600-v4.png",
                "large": "http://images-mix.netdna-ssl.com/w/300/h/300/q/85/"
                "images/graphics/33_Profile/default_user_600x600-v4.png",
                "640wx640h": "http://images-mix.netdna-ssl.com/w/640/h/640/q/85/"
                "images/graphics/33_Profile/"
                "default_user_600x600-v4.png",
                "medium_mobile": "http://images-mix.netdna-ssl.com/w/80/h/80/q/75/"
                "images/graphics/33_Profile/"
                "default_user_600x600-v4.png",
                "small": "http://images-mix.netdna-ssl.com/w/25/h/25/q/85/images/"
                "graphics/33_Profile/default_user_600x600-v4.png",
                "thumbnail": "http://images-mix.netdna-ssl.com/w/50/h/50/q/85/"
                "images/graphics/33_Profile/"
                "default_user_600x600-v4.png",
            },
            "is_current_user": True,
            "listen_count": 0,
            "updated_time": "2013-03-17T06:26:31Z",
            "following": False,
            "follower": False,
            "key": "/foobar/",
            "created_time": "2013-03-17T06:26:31Z",
            "follower_count": 0,
            "favorite_count": 0,
            "name": "foobar",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



# pyright: reportAttributeAccessIssue=false

import sys
from html.parser import HTMLParser

import requests
from httpretty import HTTPretty
from openid import oidutil

from ...backends.utils import load_backends
from ...utils import module_member, parse_qs
from ..models import TestAssociation, TestNonce, TestStorage, TestUserSocialAuth, User
from ..strategy import TestStrategy
from .base import BaseBackendTest

sys.path.insert(0, "..")


# Patch to remove the too-verbose output until a new version is released
oidutil.log = lambda *args, **kwargs: None


class FormHTMLParser(HTMLParser):
    form = {}
    inputs = {}

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag == "form":
            self.form.update(attrs)
        elif tag == "input" and "name" in attrs:
            self.inputs[attrs["name"]] = attrs["value"]


class OpenIdTest(BaseBackendTest):
    backend_path = None
    backend = None
    access_token_body = None
    user_data_body = None
    user_data_url = ""
    expected_username = ""
    settings = None
    partial_login_settings = None
    raw_complete_url = "/complete/{0}/"

    def setUp(self):
        HTTPretty.enable(allow_net_connect=False)
        Backend = module_member(self.backend_path)
        self.strategy = TestStrategy(TestStorage)
        self.complete_url = self.raw_complete_url.format(Backend.name)
        self.backend = Backend(self.strategy, redirect_uri=self.complete_url)
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_AUTHENTICATION_BACKENDS": (
                    self.backend_path,
                    "social_core.tests.backends.test_broken.BrokenBackendAuth",
                )
            }
        )
        # Force backends loading to trash PSA cache
        load_backends(
            self.strategy.get_setting("SOCIAL_AUTH_AUTHENTICATION_BACKENDS"),
            force_load=True,
        )

    def tearDown(self):
        self.strategy = None
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        HTTPretty.disable()
        HTTPretty.reset()

    def get_form_data(self, html):
        parser = FormHTMLParser()
        parser.feed(html)
        return parser.form, parser.inputs

    def openid_url(self):
        return self.backend.openid_url()

    def post_start(self):
        pass

    def do_start(self):
        HTTPretty.register_uri(
            HTTPretty.GET,
            self.openid_url(),
            status=200,
            body=self.discovery_body,
            content_type="application/xrds+xml",
        )
        start = self.backend.start()
        self.post_start()
        form, inputs = self.get_form_data(start)
        action = form.get("action")
        assert action, "The form action must be set in the test"
        HTTPretty.register_uri(
            HTTPretty.POST, action, status=200, body=self.server_response
        )
        response = requests.post(action, data=inputs)
        self.strategy.set_request_data(parse_qs(response.content), self.backend)
        HTTPretty.register_uri(
            HTTPretty.POST, form.get("action"), status=200, body="is_valid:true\n"
        )
        return self.backend.complete()



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class TumblrOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.tumblr.TumblrOAuth"
    user_data_url = "http://api.tumblr.com/v2/user/info"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "meta": {"status": 200, "msg": "OK"},
            "response": {
                "user": {
                    "following": 1,
                    "blogs": [
                        {
                            "updated": 0,
                            "description": "",
                            "drafts": 0,
                            "title": "Untitled",
                            "url": "http://foobar.tumblr.com/",
                            "messages": 0,
                            "tweet": "N",
                            "share_likes": True,
                            "posts": 0,
                            "primary": True,
                            "queue": 0,
                            "admin": True,
                            "followers": 0,
                            "ask": False,
                            "facebook": "N",
                            "type": "public",
                            "facebook_opengraph_enabled": "N",
                            "name": "foobar",
                        }
                    ],
                    "default_post_format": "html",
                    "name": "foobar",
                    "likes": 0,
                }
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Coding OAuth2 backend, docs at:
"""

from urllib.parse import urljoin

from .oauth import BaseOAuth2


class CodingOAuth2(BaseOAuth2):
    """Coding OAuth authentication backend"""

    name = "coding"
    API_URL = "https://coding.net/api/"
    AUTHORIZATION_URL = "https://coding.net/oauth_authorize.html"
    ACCESS_TOKEN_URL = "https://coding.net/api/oauth/access_token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    DEFAULT_SCOPE = ["user"]
    REDIRECT_STATE = False

    def api_url(self):
        return self.API_URL

    def get_user_details(self, response):
        """Return user details from Github account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": response.get("name"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        data = self._user_data(access_token)
        if data.get("code") != 0:
            # 获取失败
            pass
        return data.get("data")

    def _user_data(self, access_token, path=None):
        url = urljoin(self.api_url(), "account/current_user{}".format(path or ""))
        return self.get_json(url, params={"access_token": access_token})



from .oauth import BaseOAuth2


class MailChimpOAuth2(BaseOAuth2):
    """MailChimp OAuth2 authentication backend"""

    name = "mailchimp"
    AUTHORIZATION_URL = "https://login.mailchimp.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://login.mailchimp.com/oauth2/token"
    METADATA_URL = "https://login.mailchimp.com/oauth2/metadata"
    ACCESS_TOKEN_METHOD = "POST"
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    ID_KEY = "user_id"
    EXTRA_DATA = [
        ("accountname", "accountname"),
        ("api_endpoint", "api_endpoint"),
        ("role", "role"),
        ("login", "login"),
    ]

    def get_user_details(self, response):
        """Return user details from a Mailchimp metadata response"""
        return {
            "username": response["login"]["login_name"],
            "email": response["login"]["email"],
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data and datacenter information from service"""
        return self.get_json(
            self.METADATA_URL, headers={"Authorization": "OAuth " + access_token}
        )



import json

from httpretty import HTTPretty

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class AmazonOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.amazon.AmazonOAuth2"
    user_data_url = "https://api.amazon.com/user/profile"
    expected_username = "FooBar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "user_id": "amzn1.account.ABCDE1234",
            "email": "foo@bar.com",
            "name": "Foo Bar",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class AmazonOAuth2BrokenServerResponseTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.amazon.AmazonOAuth2"
    user_data_url = "https://www.amazon.com/ap/user/profile"
    expected_username = "FooBar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "Request-Id": "02GGTU7CWMNFTV3KH3J6",
            "Profile": {
                "Name": "Foo Bar",
                "CustomerId": "amzn1.account.ABCDE1234",
                "PrimaryEmail": "foo@bar.com",
            },
        }
    )

    def setUp(self):
        super().setUp()
        HTTPretty.register_uri(
            HTTPretty.GET,
            "https://api.amazon.com/user/profile",
            status=200,
            body=self.user_data_body,
            content_type="application/json",
        )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
DailyMotion OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dailymotion.html
"""

from .oauth import BaseOAuth2


class DailymotionOAuth2(BaseOAuth2):
    """Dailymotion OAuth authentication backend"""

    name = "dailymotion"
    EXTRA_DATA = [("id", "id")]
    ID_KEY = "username"
    AUTHORIZATION_URL = "https://api.dailymotion.com/oauth/authorize"
    REQUEST_TOKEN_URL = "https://api.dailymotion.com/oauth/token"
    ACCESS_TOKEN_URL = "https://api.dailymotion.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"

    def get_user_details(self, response):
        return {"username": response.get("screenname")}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "https://api.dailymotion.com/auth/", params={"access_token": access_token}
        )



"""
Mixcloud OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mixcloud.html
"""

from .oauth import BaseOAuth2


class MixcloudOAuth2(BaseOAuth2):
    name = "mixcloud"
    ID_KEY = "username"
    AUTHORIZATION_URL = "https://www.mixcloud.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://www.mixcloud.com/oauth/access_token"
    ACCESS_TOKEN_METHOD = "POST"

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(response["name"])
        return {
            "username": response["username"],
            "email": None,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://api.mixcloud.com/me/",
            params={"access_token": access_token, "alt": "json"},
        )



"""
Twitter OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/twitter-oauth2.html
    https://developer.twitter.com/en/docs/authentication/oauth-2-0/authorization-code
"""

from .oauth import BaseOAuth2PKCE


class TwitterOAuth2(BaseOAuth2PKCE):
    """Twitter OAuth2 authentication backend"""

    name = "twitter-oauth2"
    AUTHORIZATION_URL = "https://twitter.com/i/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://api.twitter.com/2/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    DEFAULT_SCOPE = ["users.read", "tweet.read"]
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    USE_BASIC_AUTH = True
    ID_KEY = "id"
    EXTRA_DATA = [
        ("id", "id"),
        ("username", "username"),
        ("fullname", "fullname"),
        ("first_name", "first_name"),
        ("last_name", "last_name"),
        ("created_at", "created_at"),
        ("verified", "verified"),
        ("verified_type", "verified_type"),
        ("proteted", "protected"),
        ("description", "description"),
        ("url", "url"),
        ("profile_image_url", "profile_image_url"),
        ("pinned_tweet_id", "pinned_tweet_id"),
        ("public_metrics", "public_metrics"),
    ]
    PKCE_DEFAULT_CODE_CHALLENGE_METHOD = "s256"
    PKCE_DEFAULT_CODE_VERIFIER_LENGTH = 32
    DEFAULT_USE_PKCE = True

    def get_user_details(self, response):
        """Return user details from Twitter account"""
        user = response
        user_id = user["id"]
        name = user["name"]
        username = user["username"]

        created_at = user.get("created_at")
        verified = user.get("verified")
        verified_type = user.get("verified_type")
        protected = user.get("protected")
        description = user.get("description")
        url = user.get("url")
        profile_image_url = user.get("profile_image_url")
        pinned_tweet_id = user.get("pinned_tweet_id")
        public_metrics = user.get("public_metrics")

        fullname, first_name, last_name = self.get_user_names(name)

        return {
            "id": user_id,
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "created_at": created_at,
            "verified": verified,
            "verified_type": verified_type,
            "protected": protected,
            "description": description,
            "url": url,
            "pinned_tweet_id": pinned_tweet_id,
            "profile_image_url": profile_image_url,
            "public_metrics": public_metrics,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        # https://developer.twitter.com/en/docs/twitter-api/users/lookup/api-reference/get-users-me
        fields = [
            "created_at",
            "description",
            "entities",
            "id",
            "location",
            "name",
            "pinned_tweet_id",
            "profile_image_url",
            "protected",
            "public_metrics",
            "url",
            "username",
            "verified",
            "verified_type",
            "withheld",
        ]
        response = self.get_json(
            "https://api.twitter.com/2/users/me",
            params={"user.fields": ",".join(fields)},
            headers={"Authorization": f"Bearer {access_token}"},
        )
        return response["data"]



"""
Coursera OAuth2 backend, docs at:
    https://tech.coursera.org/app-platform/oauth2/
"""

from .oauth import BaseOAuth2


class CourseraOAuth2(BaseOAuth2):
    """Coursera OAuth2 authentication backend"""

    name = "coursera"
    ID_KEY = "username"
    AUTHORIZATION_URL = "https://accounts.coursera.org/oauth2/v1/auth"
    ACCESS_TOKEN_URL = "https://accounts.coursera.org/oauth2/v1/token"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ","
    DEFAULT_SCOPE = ["view_profile"]

    def _get_username_from_response(self, response):
        elements = response.get("elements", [])
        for element in elements:
            if "id" in element:
                return element.get("id")

        return None

    def get_user_details(self, response):
        """Return user details from Coursera account"""
        return {"username": self._get_username_from_response(response)}

    def get_user_id(self, details, response):
        """Return a username prepared in get_user_details as uid"""
        return details.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from the service"""
        return self.get_json(
            "https://api.coursera.org/api/externalBasicProfiles.v1?q=me",
            headers=self.get_auth_header(access_token),
        )

    def get_auth_header(self, access_token):
        return {"Authorization": f"Bearer {access_token}"}



"""
Fedora OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/fedora.html
"""

from .open_id import OpenIdAuth


class FedoraOpenId(OpenIdAuth):
    name = "fedora"
    URL = "https://id.fedoraproject.org"
    USERNAME_KEY = "nickname"



# pyright: reportAttributeAccessIssue=false

import base64
import datetime
import json
import os
import sys
from calendar import timegm
from urllib.parse import urlparse

import jwt
from httpretty import HTTPretty

from social_core.backends.open_id_connect import OpenIdConnectAuth

from ...exceptions import AuthTokenError
from ...utils import parse_qs
from .oauth import BaseAuthUrlTestMixin, OAuth2Test

sys.path.insert(0, "..")


TEST_ROOT = os.path.dirname(os.path.dirname(__file__))

JWK_KEY = {
    "kty": "RSA",
    "d": "ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o"
    "_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK"
    "JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I"
    "-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ"
    "m-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4"
    "4QDSpQDCHfgcGOlHY_t5gQ",
    "e": "AQAB",
    "use": "sig",
    "kid": "testkey",
    "alg": "RS256",
    "n": "pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8"
    "H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao"
    "Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA"
    "b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g"
    "ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT"
    "vYVCvN4rBT1DQ-x0F1mo2Q",
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != "d"}


class OpenIdConnectTestMixin:
    """
    Mixin to test OpenID Connect consumers. Inheriting classes should also
    inherit OAuth2Test.
    """

    client_key = "a-key"
    client_secret = "a-secret-key"
    issuer = None  # id_token issuer
    openid_config_body = None
    key = None

    # Avoid sharing access_token_kwargs between different subclasses
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.access_token_kwargs = getattr(cls, "access_token_kwargs", {})

    def setUp(self):
        super().setUp()
        self.key = JWK_KEY.copy()
        self.public_key = JWK_PUBLIC_KEY.copy()

        assert self.openid_config_body, "openid_config_body must be set"

        HTTPretty.register_uri(
            HTTPretty.GET,
            self.backend.oidc_endpoint() + "/.well-known/openid-configuration",
            status=200,
            body=self.openid_config_body,
        )
        oidc_config = json.loads(self.openid_config_body)

        def jwks(_request, _uri, headers):
            return 200, headers, json.dumps({"keys": [self.key]})

        HTTPretty.register_uri(
            HTTPretty.GET,
            oidc_config.get("jwks_uri"),
            status=200,
            body=json.dumps({"keys": [self.public_key]}),
        )

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {
                f"SOCIAL_AUTH_{self.name}_KEY": self.client_key,
                f"SOCIAL_AUTH_{self.name}_SECRET": self.client_secret,
                f"SOCIAL_AUTH_{self.name}_ID_TOKEN_DECRYPTION_KEY": self.client_secret,
            }
        )
        return settings

    def get_id_token(
        self,
        client_key=None,
        expiration_datetime=None,
        issue_datetime=None,
        nonce=None,
        issuer=None,
    ):
        """
        Return the id_token to be added to the access token body.
        """
        return {
            "iss": issuer,
            "nonce": nonce,
            "aud": client_key,
            "azp": client_key,
            "exp": expiration_datetime,
            "iat": issue_datetime,
            "sub": "1234",
        }

    def prepare_access_token_body(
        self,
        client_key=None,
        tamper_message=False,
        expiration_datetime=None,
        kid=None,
        issue_datetime=None,
        nonce=None,
        issuer=None,
    ):
        """
        Prepares a provider access token response. Arguments:

        client_id       -- (str) OAuth ID for the client that requested
                                 authentication.
        expiration_time -- (datetime) Date and time after which the response
                                      should be considered invalid.
        """

        body = {"access_token": "foobar", "token_type": "bearer"}
        client_key = client_key or self.client_key
        now = datetime.datetime.now(datetime.timezone.utc)
        expiration_datetime = expiration_datetime or (
            now + datetime.timedelta(seconds=30)
        )
        issue_datetime = issue_datetime or now
        nonce = nonce or "a-nonce"
        issuer = issuer or self.issuer
        id_token = self.get_id_token(
            client_key,
            timegm(expiration_datetime.timetuple()),
            timegm(issue_datetime.timetuple()),
            nonce,
            issuer,
        )
        # calc at_hash
        id_token["at_hash"] = OpenIdConnectAuth.calc_at_hash("foobar", "RS256")

        body["id_token"] = jwt.encode(
            id_token,
            key=jwt.PyJWK(
                dict(self.key, iat=timegm(issue_datetime.timetuple()), nonce=nonce)  # type: ignore reportCallIssue
            ).key,
            algorithm="RS256",
            headers={"kid": kid} if kid else None,
        )

        if tamper_message:
            header, msg, sig = body["id_token"].split(".")
            id_token["sub"] = "1235"
            msg = base64.encodebytes(json.dumps(id_token).encode()).decode()
            body["id_token"] = f"{header}.{msg}.{sig}"

        return json.dumps(body)

    def authtoken_raised(self, expected_message, **access_token_kwargs):
        self.access_token_kwargs = access_token_kwargs
        with self.assertRaisesRegex(AuthTokenError, expected_message):
            self.do_login()

    def pre_complete_callback(self, start_url):
        nonce = parse_qs(urlparse(start_url).query)["nonce"]

        self.access_token_kwargs.setdefault("nonce", nonce)
        self.access_token_body = self.prepare_access_token_body(
            **self.access_token_kwargs
        )
        super().pre_complete_callback(start_url)

    def test_invalid_signature(self):
        self.authtoken_raised(
            "Token error: Signature verification failed", tamper_message=True
        )

    def test_expired_signature(self):
        expiration_datetime = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(seconds=30)
        self.authtoken_raised(
            "Token error: Signature has expired",
            expiration_datetime=expiration_datetime,
        )

    def test_invalid_issuer(self):
        self.authtoken_raised("Token error: Invalid issuer", issuer="someone-else")

    def test_invalid_audience(self):
        self.authtoken_raised(
            "Token error: Invalid audience", client_key="someone-else"
        )

    def test_invalid_issue_time(self):
        expiration_datetime = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(seconds=self.backend.ID_TOKEN_MAX_AGE * 2)
        self.authtoken_raised(
            "Token error: Incorrect id_token: iat", issue_datetime=expiration_datetime
        )

    def test_invalid_nonce(self):
        self.authtoken_raised(
            "Token error: Incorrect id_token: nonce",
            nonce="something-wrong",
            kid="testkey",
        )

    def test_invalid_kid(self):
        self.authtoken_raised(
            "Token error: Signature verification failed", kid="doesnotexist"
        )


class BaseOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.open_id_connect.OpenIdConnectAuth"
    issuer = "https://example.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://example.com",
            "authorization_endpoint": "https://example.com/oidc/auth",
            "token_endpoint": "https://example.com/oidc/token",
            "userinfo_endpoint": "https://example.com/oidc/userinfo",
            "revocation_endpoint": "https://example.com/oidc/revoke",
            "jwks_uri": "https://example.com/oidc/certs",
        }
    )

    expected_username = "cartman"

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {
                "SOCIAL_AUTH_OIDC_OIDC_ENDPOINT": "https://example.com/oidc",
            }
        )
        return settings

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        HTTPretty.register_uri(
            "GET",
            uri=self.backend.userinfo_url(),
            status=200,
            body=json.dumps({"preferred_username": self.expected_username}),
            content_type="text/json",
        )

    def test_everything_works(self):
        self.do_login()


class ExampleOpenIdConnectAuth(OpenIdConnectAuth):
    name = "example123"
    OIDC_ENDPOINT = "https://example.com/oidc"


class ExampleOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = (
        "social_core.tests.backends.test_open_id_connect.ExampleOpenIdConnectAuth"
    )
    issuer = "https://example.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://example.com",
            "authorization_endpoint": "https://example.com/oidc/auth",
            "token_endpoint": "https://example.com/oidc/token",
            "userinfo_endpoint": "https://example.com/oidc/userinfo",
            "revocation_endpoint": "https://example.com/oidc/revoke",
            "jwks_uri": "https://example.com/oidc/certs",
        }
    )

    expected_username = "cartman"

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        HTTPretty.register_uri(
            "GET",
            uri=self.backend.userinfo_url(),
            status=200,
            body=json.dumps({"preferred_username": self.expected_username}),
            content_type="text/json",
        )

    def test_everything_works(self):
        self.do_login()



import json

from httpretty import HTTPretty

from ...exceptions import AuthFailed
from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class GithubEnterpriseOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.github_enterprise.GithubEnterpriseOAuth2"
    user_data_url = "https://www.example.com/api/v3/user"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://www.example.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://www.example.com/api/v3/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://www.example.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://www.example.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL": "https://www.example.com/api/v3"}
        )
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL": "https://www.example.com/api/v3"}
        )
        self.do_partial_pipeline()


class GithubEnterpriseOAuth2NoEmailTest(GithubEnterpriseOAuth2Test):
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://www.example.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://www.example.com/api/v3/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://www.example.com/blog",
            "location": "San Francisco",
            "email": "",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://www.example.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL": "https://www.example.com/api/v3"}
        )
        url = "https://www.example.com/api/v3/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps(["foo@bar.com"]),
            content_type="application/json",
        )
        self.do_login()

    def test_login_next_format(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL": "https://www.example.com/api/v3"}
        )
        url = "https://www.example.com/api/v3/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps([{"email": "foo@bar.com"}]),
            content_type="application/json",
        )
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL": "https://www.example.com/api/v3"}
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            "https://www.example.com/api/v3/user/emails",
            status=200,
            body=json.dumps([{"email": "foo@bar.com"}]),
            content_type="application/json",
        )
        self.do_partial_pipeline()


class GithubEnterpriseOrganizationOAuth2Test(GithubEnterpriseOAuth2Test):
    backend_path = (
        "social_core.backends.github_enterprise.GithubEnterpriseOrganizationOAuth2"
    )

    def auth_handlers(self, start_url):
        url = "https://www.example.com/api/v3/orgs/foobar/members/foobar"
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME": "foobar"})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME": "foobar"})
        self.do_partial_pipeline()


class GithubEnterpriseOrganizationOAuth2FailTest(GithubEnterpriseOAuth2Test):
    backend_path = (
        "social_core.backends.github_enterprise.GithubEnterpriseOrganizationOAuth2"
    )

    def auth_handlers(self, start_url):
        url = "https://www.example.com/api/v3/orgs/foobar/members/foobar"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()


class GithubEnterpriseTeamOAuth2Test(GithubEnterpriseOAuth2Test):
    backend_path = "social_core.backends.github_enterprise.GithubEnterpriseTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://www.example.com/api/v3/teams/123/members/foobar"
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID": "123"})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID": "123"})
        self.do_partial_pipeline()


class GithubEnterpriseTeamOAuth2FailTest(GithubEnterpriseOAuth2Test):
    backend_path = "social_core.backends.github_enterprise.GithubEnterpriseTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://www.example.com/api/v3/teams/123/members/foobar"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL": "https://www.example.com"}
        )
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL": "https://www.example.com/api/v3"
            }
        )
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()



"""
Yammer OAuth2 production and staging backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/yammer.html
"""

from .oauth import BaseOAuth2


class YammerOAuth2(BaseOAuth2):
    name = "yammer"
    AUTHORIZATION_URL = "https://www.yammer.com/dialog/oauth"
    ACCESS_TOKEN_URL = "https://www.yammer.com/oauth2/access_token"
    EXTRA_DATA = [("id", "id"), ("expires", "expires"), ("mugshot_url", "mugshot_url")]

    def get_user_id(self, details, response):
        return response["user"]["id"]

    def get_user_details(self, response):
        username = response["user"]["name"]
        fullname, first_name, last_name = self.get_user_names(
            fullname=response["user"]["full_name"],
            first_name=response["user"]["first_name"],
            last_name=response["user"]["last_name"],
        )
        email = response["user"]["contact"]["email_addresses"][0]["address"]
        mugshot_url = response["user"]["mugshot_url"]
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "picture_url": mugshot_url,
        }


class YammerStagingOAuth2(YammerOAuth2):
    name = "yammer-staging"
    AUTHORIZATION_URL = "https://www.staging.yammer.com/dialog/oauth"
    ACCESS_TOKEN_URL = "https://www.staging.yammer.com/oauth2/access_token"
    REQUEST_TOKEN_URL = "https://www.staging.yammer.com/oauth2/request_token"



"""
Foursquare OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/foursquare.html
"""

from .oauth import BaseOAuth2


class FoursquareOAuth2(BaseOAuth2):
    name = "foursquare"
    AUTHORIZATION_URL = "https://foursquare.com/oauth2/authenticate"
    ACCESS_TOKEN_URL = "https://foursquare.com/oauth2/access_token"
    ACCESS_TOKEN_METHOD = "POST"
    API_VERSION = "20140128"

    def get_user_id(self, details, response):
        return response["response"]["user"]["id"]

    def get_user_details(self, response):
        """Return user details from Foursquare account"""
        info = response["response"]["user"]
        email = info["contact"]["email"]
        fullname, first_name, last_name = self.get_user_names(
            first_name=info.get("firstName", ""), last_name=info.get("lastName", "")
        )
        return {
            "username": first_name + " " + last_name,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.foursquare.com/v2/users/self",
            params={"oauth_token": access_token, "v": self.API_VERSION},
        )



from .oauth import BaseOAuth2PKCE


class EtsyOAuth2(BaseOAuth2PKCE):
    name = "etsy"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://www.etsy.com/oauth/connect"
    ACCESS_TOKEN_URL = "https://api.etsy.com/v3/public/oauth/token"
    REFRESH_TOKEN_URL = "https://api.etsy.com/v3/public/oauth/token"
    PKCE_DEFAULT_CODE_CHALLENGE_METHOD = "S256"
    ACCESS_TOKEN_METHOD = "POST"
    REQUEST_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
        ("token_type", "token_type"),
        ("access_token", "access_token"),
        # User Data Fields
        ("primary_email", "primary_email"),
        ("first_name", "first_name"),
        ("last_name", "last_name"),
        ("image_url_75x75", "image_url_75x75"),
    ]

    def user_data(self, access_token, *args, **kwargs) -> dict:
        client_id, _ = self.get_key_and_secret()
        user_id = access_token.split(".")[0]
        headers = {"Authorization": f"Bearer {access_token}", "x-api-key": client_id}
        return self.get_json(
            url=f"https://openapi.etsy.com/v3/application/users/{user_id}",
            headers=headers,
        )

    def get_user_details(self, response):
        return {
            "user_id": response["user_id"],
            "first_name": response["first_name"],
            "last_name": response["last_name"],
            "email": response["primary_email"],
            "image_url_75x75": response["image_url_75x75"],
            "username": str(response["user_id"]),
        }



from .oauth import BaseOAuth2


class UniverseOAuth2(BaseOAuth2):
    """Universe Ticketing OAuth2 authentication backend"""

    name = "universe"
    AUTHORIZATION_URL = "https://www.universe.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://www.universe.com/oauth/token"
    BASE_API_URL = "https://www.universe.com/api"
    USER_INFO_URL = BASE_API_URL + "/v2/current_user"
    ACCESS_TOKEN_METHOD = "POST"
    STATE_PARAMETER = True
    REDIRECT_STATE = True
    EXTRA_DATA = [
        ("id", "id"),
        ("slug", "slug"),
        ("created_at", "created_at"),
        ("updated_at", "updated_at"),
    ]

    def get_user_id(self, details, response):
        return response["current_user"][self.ID_KEY]

    def get_user_details(self, response):
        """Return user details from a Universe account"""
        # Start with the user data as it was returned
        user_details = response["current_user"]
        user_details["username"] = user_details["email"]
        return user_details

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            self.USER_INFO_URL, headers={"Authorization": f"Bearer {access_token}"}
        )



import json
import re
import sys
import unittest
from os import path
from unittest.mock import patch
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import requests
from httpretty import HTTPretty

try:
    from onelogin.saml2.utils import (  # type: ignore reportMissingImports
        OneLogin_Saml2_Utils,
    )

    SAML_MODULE_ENABLED = True
except ImportError:
    SAML_MODULE_ENABLED = False

from ...exceptions import AuthMissingParameter
from .base import BaseBackendTest

DATA_DIR = path.join(path.dirname(__file__), "data")


@unittest.skipIf(
    "__pypy__" in sys.builtin_module_names, "dm.xmlsec not compatible with pypy"
)
@unittest.skipUnless(SAML_MODULE_ENABLED, "Only run if onelogin.saml2 is installed")
class SAMLTest(BaseBackendTest):
    backend_path = "social_core.backends.saml.SAMLAuth"
    expected_username = "myself"
    response_fixture = "saml_response.txt"

    def extra_settings(self):
        name = path.join(DATA_DIR, "saml_config.json")
        with open(name) as config_file:
            config_str = config_file.read()
        return json.loads(config_str)

    def setUp(self):
        """Patch the time so that we can replay canned
        request/response pairs"""
        super().setUp()

        @staticmethod
        def fixed_time():
            return OneLogin_Saml2_Utils.parse_SAML_to_time("2015-05-09T03:57:22Z")

        now_patch = patch.object(OneLogin_Saml2_Utils, "now", fixed_time)
        now_patch.start()
        self.addCleanup(now_patch.stop)

    def install_http_intercepts(self, start_url, return_url):
        # When we request start_url
        # (https://idp.testshib.org/idp/profile/SAML2/Redirect/SSO...)
        # we will eventually get a redirect back, with SAML assertion
        # data in the query string.  A pre-recorded correct response
        # is kept in this .txt file:
        name = path.join(DATA_DIR, self.response_fixture)
        with open(name) as response_file:
            response_url = response_file.read()
        HTTPretty.register_uri(
            HTTPretty.GET, start_url, status=301, location=response_url
        )
        HTTPretty.register_uri(HTTPretty.GET, return_url, status=200, body="foobar")

    def do_start(self):
        start_url = self.backend.start().url
        # Modify the start URL to make the SAML request consistent
        # from test to test:
        start_url = self.modify_start_url(start_url)
        # If the SAML Identity Provider recognizes the user, we will
        # be redirected back to:
        return_url = self.backend.redirect_uri
        self.install_http_intercepts(start_url, return_url)
        response = requests.get(start_url)
        self.assertTrue(response.url.startswith(return_url))
        self.assertEqual(response.text, "foobar")
        query_values = {
            k: v[0] for k, v in parse_qs(urlparse(response.url).query).items()
        }
        self.assertNotIn(" ", query_values["SAMLResponse"])
        self.strategy.set_request_data(query_values, self.backend)
        return self.backend.complete()

    def test_metadata_generation(self):
        """Test that we can generate the metadata without error"""
        xml, errors = self.backend.generate_metadata_xml()
        self.assertEqual(len(errors), 0)
        self.assertEqual(xml.decode()[0], "<")

    def test_login_with_next_url(self):
        """
        Test that we login and then redirect to the "next" URL.
        """
        # pretend we've started with a URL like /login/saml/?idp=testshib&next=/foo/bar
        self.strategy.set_request_data(
            {"idp": "testshib", "next": "/foo/bar"}, self.backend
        )
        self.do_login()
        # The core `do_complete` action assumes the "next" URL is stored in session state or the request data.
        self.assertEqual(self.strategy.session_get("next"), "/foo/bar")

    def test_login_no_next_url(self):
        """
        Test that we handle "next" being omitted from the request data and RelayState.
        """
        self.response_fixture = "saml_response_no_next_url.txt"

        # pretend we've started with a URL like /login/saml/?idp=testshib
        self.strategy.set_request_data({"idp": "testshib"}, self.backend)
        self.do_login()
        self.assertEqual(self.strategy.session_get("next"), None)

    def test_login_with_legacy_relay_state(self):
        """
        Test that we handle legacy RelayState (i.e. just the IDP name, not a JSON object).

        This is the form that RelayState had in prior versions of this library. It should be supported for backwards
        compatibility.
        """
        self.response_fixture = "saml_response_legacy.txt"

        self.strategy.set_request_data({"idp": "testshib"}, self.backend)
        self.do_login()

    def test_login_no_idp_in_initial_request(self):
        """
        Logging in without an idp param should raise AuthMissingParameter
        """
        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def test_login_no_idp_in_saml_response(self):
        """
        The RelayState should always contain a JSON object with an "idp" key, or be just the IDP name as a string.
        This tests that an exception is raised if it is a JSON object, but is missing the "idp" key.
        """
        self.response_fixture = "saml_response_no_idp_name.txt"

        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def modify_start_url(self, start_url):
        """
        Given a SAML redirect URL, parse it and change the ID to
        a consistent value, so the request is always identical.
        """
        # Parse the SAML Request URL to get the XML being sent to TestShib
        url_parts = urlparse(start_url)
        query = {k: v[0] for (k, v) in parse_qs(url_parts.query).items()}
        xml = OneLogin_Saml2_Utils.decode_base64_and_inflate(query["SAMLRequest"])
        # Modify the XML:
        xml = xml.decode()
        xml, changed = re.subn(r'ID="[^"]+"', 'ID="TEST_ID"', xml)
        self.assertEqual(changed, 1)
        # Update the URL to use the modified query string:
        query["SAMLRequest"] = OneLogin_Saml2_Utils.deflate_and_base64_encode(xml)
        url_parts = list(url_parts)
        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)



"""
Gitea OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/gitea.html
"""

from .oauth import BaseOAuth2


class GiteaOAuth2(BaseOAuth2):
    """Gitea OAuth authentication backend"""

    name = "gitea"
    API_URL = "https://gitea.com"
    AUTHORIZATION_URL = "https://gitea.com/login/oauth/authorize"
    ACCESS_TOKEN_URL = "https://gitea.com/login/oauth/access_token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    EXTRA_DATA = [
        ("id", "id"),
        ("expires_in", "expires"),
        ("refresh_token", "refresh_token"),
    ]

    def api_url(self, path):
        api_url = self.setting("API_URL") or self.API_URL
        return "{}{}".format(api_url.rstrip("/"), path)

    def authorization_url(self):
        return self.api_url("/login/oauth/authorize")

    def access_token_url(self):
        return self.api_url("/login/oauth/access_token")

    def get_user_details(self, response):
        """Return user details from Gitea account"""
        fullname, first_name, last_name = self.get_user_names(response.get("fullname"))
        return {
            "username": response.get("login"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            self.api_url("/api/v1/user"), params={"access_token": access_token}
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class QuizletOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.quizlet.QuizletOAuth2"
    expected_username = "foo_bar"

    access_token_body = json.dumps(
        {
            "access_token": "EE1IDxytP04tJ767GbjH7ED9PpGmYvL",
            "token_type": "Bearer",
            "expires_in": 3600,
            "scope": "read",
            "user_id": "foo_bar",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Box.net OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/box.html
"""

from .oauth import BaseOAuth2


class BoxOAuth2(BaseOAuth2):
    """Box.net OAuth authentication backend"""

    name = "box"
    AUTHORIZATION_URL = "https://www.box.com/api/oauth2/authorize"
    ACCESS_TOKEN_METHOD = "POST"
    ACCESS_TOKEN_URL = "https://www.box.com/api/oauth2/token"
    REVOKE_TOKEN_URL = "https://www.box.com/api/oauth2/revoke"
    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("id", "id"),
        ("expires", "expires"),
    ]

    def do_auth(self, access_token, response=None, *args, **kwargs):
        response = response or {}
        data = self.user_data(access_token)

        data["access_token"] = response.get("access_token")
        data["refresh_token"] = response.get("refresh_token")
        data["expires"] = response.get("expires_in")
        kwargs.update({"backend": self, "response": data})
        return self.strategy.authenticate(*args, **kwargs)

    def get_user_details(self, response):
        """Return user details Box.net account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": response.get("login"),
            "email": response.get("login") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        params = self.setting("PROFILE_EXTRA_PARAMS", {})
        params["access_token"] = access_token
        return self.get_json("https://api.box.com/2.0/users/me", params=params)

    def refresh_token(self, token, *args, **kwargs):
        params = self.refresh_token_params(token, *args, **kwargs)
        request = self.request(
            self.REFRESH_TOKEN_URL or self.ACCESS_TOKEN_URL,
            data=params,
            headers=self.auth_headers(),
            method="POST",
        )
        return self.process_refresh_token_response(request, *args, **kwargs)



"""
Pinterest OAuth2 backend, docs at:
    https://developers.pinterest.com/docs/api/authentication/
"""

from .oauth import BaseOAuth2


class PinterestOAuth2(BaseOAuth2):
    name = "pinterest"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://api.pinterest.com/oauth/"
    ACCESS_TOKEN_URL = "https://api.pinterest.com/v1/oauth/token"
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = "POST"

    def user_data(self, access_token, *args, **kwargs):
        response = self.get_json(
            "https://api.pinterest.com/v1/me/", params={"access_token": access_token}
        )

        if "data" in response:
            username = response["data"]["url"].strip("/").split("/")[-1]
            response = {
                "user_id": response["data"]["id"],
                "first_name": response["data"]["first_name"],
                "last_name": response["data"]["last_name"],
                "username": username,
            }
        return response

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            first_name=response["first_name"], last_name=response["last_name"]
        )

        return {
            "username": response.get("username"),
            "email": None,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



"""
Yahoo OpenId, OAuth1 and OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/yahoo.html
"""

from requests.auth import HTTPBasicAuth

from ..utils import handle_http_errors
from .oauth import BaseOAuth1, BaseOAuth2


class YahooOAuth(BaseOAuth1):
    """Yahoo OAuth authentication backend. DEPRECATED"""

    name = "yahoo-oauth"
    ID_KEY = "guid"
    AUTHORIZATION_URL = "https://api.login.yahoo.com/oauth/v2/request_auth"
    REQUEST_TOKEN_URL = "https://api.login.yahoo.com/oauth/v2/get_request_token"
    ACCESS_TOKEN_URL = "https://api.login.yahoo.com/oauth/v2/get_token"
    EXTRA_DATA = [
        ("guid", "id"),
        ("access_token", "access_token"),
        ("expires", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details from Yahoo Profile"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("givenName"), last_name=response.get("familyName")
        )
        emails = [email for email in response.get("emails", []) if email.get("handle")]
        emails.sort(key=lambda e: e.get("primary", False), reverse=True)
        return {
            "username": response.get("nickname"),
            "email": emails[0]["handle"] if emails else "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = "https://social.yahooapis.com/v1/user/{0}/profile?format=json"
        return self.get_json(
            url.format(self._get_guid(access_token)), auth=self.oauth_auth(access_token)
        )["profile"]

    def _get_guid(self, access_token):
        """
        Because you have to provide GUID for every API request it's also
        returned during one of OAuth calls
        """
        return self.get_json(
            "https://social.yahooapis.com/v1/me/guid?format=json",
            auth=self.oauth_auth(access_token),
        )["guid"]["value"]


class YahooOAuth2(BaseOAuth2):
    """Yahoo OAuth2 authentication backend"""

    name = "yahoo-oauth2"
    ID_KEY = "sub"
    AUTHORIZATION_URL = "https://api.login.yahoo.com/oauth2/request_auth"
    ACCESS_TOKEN_URL = "https://api.login.yahoo.com/oauth2/get_token"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [
        ("sub", "id"),
        ("access_token", "access_token"),
        ("expires_in", "expires"),
        ("refresh_token", "refresh_token"),
        ("token_type", "token_type"),
    ]

    def get_user_names(self, first_name, last_name):  # type: ignore[reportIncompatibleMethodOverride]
        if first_name or last_name:
            return f"{first_name} {last_name}", first_name, last_name
        return None, None, None

    def get_user_details(self, response):
        """
        Return user details from Yahoo Profile.
        To Get user email you need the profile private read permission.
        """

        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("given_name"), last_name=response.get("family_name")
        )

        email = response.get("email")
        return {
            "username": response.get("nickname"),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""

        url = "https://api.login.yahoo.com/openid/v1/userinfo"

        return self.get_json(
            url, headers={"Authorization": f"Bearer {access_token}"}, method="GET"
        )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            auth=HTTPBasicAuth(*self.get_key_and_secret()),
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )

    def refresh_token_params(self, token, *args, **kwargs):
        return {
            "refresh_token": token,
            "grant_type": "refresh_token",
            "redirect_uri": "oob",  # out of bounds
        }

    def refresh_token(self, token, *args, **kwargs):
        params = self.refresh_token_params(token, *args, **kwargs)
        url = self.REFRESH_TOKEN_URL or self.ACCESS_TOKEN_URL
        method = self.REFRESH_TOKEN_METHOD
        key = "params" if method == "GET" else "data"
        request_args = {"headers": self.auth_headers(), "method": method, key: params}
        request = self.request(
            url, auth=HTTPBasicAuth(*self.get_key_and_secret()), **request_args
        )
        return self.process_refresh_token_response(request, *args, **kwargs)

    def auth_complete_params(self, state=None):
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "redirect_uri": self.get_redirect_uri(state),
        }



import functools
import hmac
import logging
import re
import sys
import time
import unicodedata
from urllib.parse import parse_qs as battery_parse_qs
from urllib.parse import unquote, urlencode, urlparse, urlunparse

import requests

import social_core

from .exceptions import AuthCanceled, AuthForbidden, AuthUnreachableProvider

SETTING_PREFIX = "SOCIAL_AUTH"

PARTIAL_TOKEN_SESSION_NAME = "partial_pipeline_token"


social_logger = logging.getLogger("social")


def import_module(name):
    __import__(name)
    return sys.modules[name]


def module_member(name):
    mod, member = name.rsplit(".", 1)
    module = import_module(mod)
    return getattr(module, member)


def user_agent():
    """Builds a simple User-Agent string to send in requests"""
    return "social-auth-" + social_core.__version__


def url_add_parameters(url, params, _unquote_query=False):
    """Adds parameters to URL, parameter will be repeated if already present"""
    if params:
        fragments = list(urlparse(url))
        value = parse_qs(fragments[4])
        value.update(params)
        fragments[4] = urlencode(value)
        if _unquote_query:
            fragments[4] = unquote(fragments[4])
        url = urlunparse(fragments)
    return url


def to_setting_name(*names):
    return "_".join([name.upper().replace("-", "_") for name in names if name])


def setting_name(*names):
    return to_setting_name(*((SETTING_PREFIX, *names)))


def sanitize_redirect(hosts, redirect_to):
    """
    Given a list of hostnames and an untrusted URL to redirect to,
    this method tests it to make sure it isn't garbage/harmful
    and returns it, else returns None, similar as how's it done
    on django.contrib.auth.views.
    """
    # Avoid redirect on evil URLs like ///evil.com
    if (
        not redirect_to
        or not hasattr(redirect_to, "startswith")
        or redirect_to.startswith("///")
    ):
        return None

    try:
        # Don't redirect to a host that's not in the list
        netloc = urlparse(redirect_to)[1] or hosts[0]
    except (TypeError, AttributeError):
        pass
    else:
        if netloc in hosts:
            return redirect_to


def user_is_authenticated(user):
    if user and hasattr(user, "is_authenticated"):
        if callable(user.is_authenticated):
            authenticated = user.is_authenticated()
        else:
            authenticated = user.is_authenticated
    elif user:
        authenticated = True
    else:
        authenticated = False
    return authenticated


def user_is_active(user):
    if user and hasattr(user, "is_active"):
        is_active = user.is_active() if callable(user.is_active) else user.is_active
    elif user:
        is_active = True
    else:
        is_active = False
    return is_active


# This slugify version was borrowed from django revision a61dbd6
def slugify(value):
    """Converts to lowercase, removes non-word characters (alphanumerics
    and underscores) and converts spaces to hyphens. Also strips leading
    and trailing whitespace."""
    value = (
        unicodedata.normalize("NFKD", str(value))
        .encode("ascii", "ignore")
        .decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


def first(func, items):
    """Return the first item in the list for what func returns True"""
    for item in items:
        if func(item):
            return item
    return None


def parse_qs(value):
    """Like urlparse.parse_qs but transform list values to single items"""
    return drop_lists(battery_parse_qs(value))


def drop_lists(value):
    out = {}
    for key, val in value.items():
        val = val[0]
        if isinstance(key, bytes):
            key = str(key, "utf-8")
        if isinstance(val, bytes):
            val = str(val, "utf-8")
        out[key] = val
    return out


def partial_pipeline_data(backend, user=None, partial_token=None, *args, **kwargs):
    request_data = backend.strategy.request_data()

    partial_argument_name = backend.setting(
        "PARTIAL_PIPELINE_TOKEN_NAME", "partial_token"
    )
    partial_token = (
        partial_token
        or request_data.get(partial_argument_name)
        or backend.strategy.session_get(PARTIAL_TOKEN_SESSION_NAME, None)
    )

    if partial_token:
        partial = backend.strategy.partial_load(partial_token)
        partial_matches_request = False

        if partial and partial.backend == backend.name:
            partial_matches_request = True

            # Normally when resuming a pipeline, request_data will be empty. We
            # only need to check for a uid match if new data was provided (i.e.
            # if current request specifies the ID_KEY).
            if backend.ID_KEY in request_data:
                id_from_partial = partial.kwargs.get("uid")
                id_from_request = request_data.get(backend.ID_KEY)

                if id_from_partial != id_from_request:
                    partial_matches_request = False

        if partial_matches_request:
            if user:  # don't update user if it's None
                kwargs.setdefault("user", user)
            kwargs.setdefault("request", request_data)
            partial.extend_kwargs(kwargs)
            return partial
        backend.strategy.clean_partial_pipeline(partial_token)
        return None
    return None


def build_absolute_uri(host_url, path=None):
    """Build absolute URI with given (optional) path"""
    path = path or ""
    if path.startswith(("http://", "https://")):
        return path
    if host_url.endswith("/") and path.startswith("/"):
        path = path[1:]
    return host_url + path


def constant_time_compare(val1, val2):
    """Compare two values and prevent timing attacks for cryptographic use."""
    if isinstance(val1, str):
        val1 = val1.encode("utf-8")
    if isinstance(val2, str):
        val2 = val2.encode("utf-8")
    return hmac.compare_digest(val1, val2)


def is_url(value):
    return value and (value.startswith(("http://", "https://", "/")))


def setting_url(backend, *names):
    for name in names:
        if is_url(name):
            return name
        value = backend.setting(name)
        if is_url(value):
            return value
    return None


def handle_http_errors(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as err:
            social_logger.exception(
                "Request failed with %d: %s",
                err.response.status_code,
                err.response.text,
            )

            if err.response.status_code == 400:
                raise AuthCanceled(args[0], response=err.response)
            if err.response.status_code == 401:
                raise AuthForbidden(args[0])
            if err.response.status_code == 503:
                raise AuthUnreachableProvider(args[0])
            raise

    return wrapper


def append_slash(url):
    """Make sure we append a slash at the end of the URL otherwise we
    have issues with urljoin Example:
    >>> urlparse.urljoin('http://www.example.com/api/v3', 'user/1/')
    'http://www.example.com/api/user/1/'
    """
    if url and not url.endswith("/"):
        url = f"{url}/"
    return url


def get_strategy(strategy, storage, *args, **kwargs):
    Strategy = module_member(strategy)
    Storage = module_member(storage)
    return Strategy(Storage, *args, **kwargs)


class cache:
    """
    Cache decorator that caches the return value of a method for a
    specified time.

    It maintains a cache per class, so subclasses have a different cache entry
    for the same cached method.

    Does not work for methods with arguments.
    """

    def __init__(self, ttl):
        self.ttl = ttl
        self.cache = {}

    def __call__(self, fn):
        def wrapped(this):
            now = time.time()
            last_updated = None
            cached_value = None
            if this.__class__ in self.cache:
                last_updated, cached_value = self.cache[this.__class__]

            # ignoring this type issue is safe; if cached_value is returned, last_updated
            # is also set, but the type checker doesn't know it.
            if not cached_value or now - last_updated > self.ttl:  # type: ignore[reportOperatorIssue]
                try:
                    cached_value = fn(this)
                    self.cache[this.__class__] = (now, cached_value)
                except Exception:
                    # Use previously cached value when call fails, if available
                    if not cached_value:
                        raise
            return cached_value

        wrapped.invalidate = self._invalidate  # type: ignore[reportFunctionMemberAccess]
        return wrapped

    def _invalidate(self):
        self.cache.clear()



# noqa: A005
"""
Legacy Email backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/email.html
"""

from .legacy import LegacyAuth


class EmailAuth(LegacyAuth):
    name = "email"
    ID_KEY = "email"
    REQUIRES_EMAIL_VALIDATION = True
    EXTRA_DATA = ["email"]



from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .test_open_id_connect import OpenIdConnectTestMixin


class EInfraCZOpenIdConnectTest(
    OpenIdConnectTestMixin, OAuth2Test, BaseAuthUrlTestMixin
):
    backend_path = "social_core.backends.einfracz.EInfraCZOpenIdConnect"
    issuer = "https://login.e-infra.cz/oidc/"
    openid_config_body = """
    {
    "request_parameter_supported":true,
    "claims_parameter_supported":false,
    "introspection_endpoint":"https://login.e-infra.cz/oidc/introspect",
    "scopes_supported":[
        "openid",
        "profile",
        "email",
        "address",
        "phone",
        "offline_access"
    ],
    "issuer":"https://login.e-infra.cz/oidc/",
    "userinfo_encryption_enc_values_supported":[
        "XC20P",
        "A256CBC+HS512",
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128CBC+HS256"
    ],
    "id_token_encryption_enc_values_supported":[
        "XC20P",
        "A256CBC+HS512",
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128CBC+HS256"
    ],
    "authorization_endpoint":"https://login.e-infra.cz/oidc/authorize",
    "service_documentation":"https://login.e-infra.cz/oidc/about",
    "request_object_encryption_enc_values_supported":[
        "XC20P",
        "A256CBC+HS512",
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128CBC+HS256"
    ],
    "device_authorization_endpoint":"https://login.e-infra.cz/oidc/devicecode",
    "userinfo_signing_alg_values_supported":[
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512"
    ],
    "claims_supported":[
        "sub",
        "name",
        "preferred_username",
        "given_name",
        "family_name",
        "middle_name",
        "nickname",
        "profile",
        "picture",
        "website",
        "gender",
        "zoneinfo",
        "locale",
        "updated_at",
        "birthdate",
        "email",
        "email_verified",
        "phone_number",
        "phone_number_verified",
        "address"
    ],
    "claim_types_supported":[
        "normal"
    ],
    "op_policy_uri":"https://login.e-infra.cz/oidc/about",
    "token_endpoint_auth_methods_supported":[
        "client_secret_post",
        "client_secret_basic",
        "client_secret_jwt",
        "private_key_jwt",
        "none"
    ],
    "token_endpoint":"https://login.e-infra.cz/oidc/token",
    "response_types_supported":[
        "code",
        "token"
    ],
    "request_uri_parameter_supported":false,
    "userinfo_encryption_alg_values_supported":[
        "RSA-OAEP-512",
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5",
        "RSA-OAEP-384"
    ],
    "grant_types_supported":[
        "authorization_code",
        "implicit",
        "urn:ietf:params:oauth:grant-type:jwt-bearer",
        "client_credentials",
        "urn:ietf:params:oauth:grant_type:redelegate",
        "urn:ietf:params:oauth:grant-type:device_code",
        "refresh_token"
    ],
    "end_session_endpoint":"https://login.e-infra.cz/oidc/endsession",
    "revocation_endpoint":"https://login.e-infra.cz/oidc/revoke",
    "userinfo_endpoint":"https://login.e-infra.cz/oidc/userinfo",
    "token_endpoint_auth_signing_alg_values_supported":[
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512"
    ],
    "op_tos_uri":"https://login.e-infra.cz/oidc/about",
    "require_request_uri_registration":false,
    "code_challenge_methods_supported":[
        "plain",
        "S256"
    ],
    "id_token_encryption_alg_values_supported":[
        "RSA-OAEP-512",
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5",
        "RSA-OAEP-384"
    ],
    "jwks_uri":"https://login.e-infra.cz/oidc/jwk",
    "subject_types_supported":[
        "public",
        "pairwise"
    ],
    "id_token_signing_alg_values_supported":[
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512",
        "none"
    ],
    "registration_endpoint":"https://login.e-infra.cz/oidc/register",
    "request_object_signing_alg_values_supported":[
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512"
    ],
    "request_object_encryption_alg_values_supported":[
        "RSA-OAEP-512",
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5",
        "RSA-OAEP-384"
    ]
}
    """



from social_core.backends.oauth import BaseOAuth2


class CognitoOAuth2(BaseOAuth2):
    name = "cognito"
    ID_KEY = "username"
    DEFAULT_SCOPE = ["openid", "profile", "email"]
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False

    def user_pool_domain(self):
        return self.setting("POOL_DOMAIN")

    def authorization_url(self):
        return f"{self.user_pool_domain()}/login"

    def access_token_url(self):
        return f"{self.user_pool_domain()}/oauth2/token"

    def user_data_url(self):
        return f"{self.user_pool_domain()}/oauth2/userInfo"

    def get_user_details(self, response):
        """Return user details from their cognito pool account"""
        first_name = response.get("given_name") or ""
        last_name = response.get("family_name") or ""
        fullname, first_name, last_name = self.get_user_names(
            first_name=first_name,
            last_name=last_name,
        )
        return {
            "username": response.get("username") or response.get("email"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from cognito."""
        response = self.get_json(
            url=self.user_data_url(),
            headers={"Authorization": f"Bearer {access_token}"},
        )

        return {
            "given_name": response.get("given_name"),
            "family_name": response.get("family_name"),
            "username": response.get("username"),
            "email": response.get("email"),
        }



"""
SurveyMonkey OAuth2 backend, docs at:
    https://developer.surveymonkey.com/api/v3/#authentication
"""

from .oauth import BaseOAuth2


class SurveyMonkeyOAuth2(BaseOAuth2):
    """SurveyMonkey OAuth2 authentication backend"""

    name = "surveymonkey"
    AUTHORIZATION_URL = "https://api.surveymonkey.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.surveymonkey.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    USER_DATA_URL = "/v3/users/me"
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("access_url", "access_url"),
    ]

    def get_user_details(self, response):
        """Return user details from a SurveyMonkey /users/me response"""
        response["name"] = response["first_name"] + " " + response["last_name"]
        return response

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data information from service"""
        base_url = kwargs["response"]["access_url"]
        return self.get_json(
            base_url + self.USER_DATA_URL,
            headers={"Authorization": "bearer " + access_token},
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class KakaoOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.kakao.KakaoOAuth2"
    user_data_url = "https://kapi.kakao.com/v2/user/me"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar"})
    user_data_body = json.dumps(
        {
            "id": "101010101",
            "properties": {
                "nickname": "foobar",
                "thumbnail_image": "http://mud-kage.kakao.co.kr/14/dn/btqbh1AKmRf/"
                "ujlHpQhxtMSbhKrBisrxe1/o.jpg",
                "profile_image": "http://mud-kage.kakao.co.kr/14/dn/btqbjCnl06Q/"
                "wbMJSVAUZB7lzSImgGdsoK/o.jpg",
            },
            "kakao_account": {"email": "a"},
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class TaobaoOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.taobao.TAOBAOAuth"
    user_data_url = "https://eco.taobao.com/router/rest"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "w2_expires_in": 0,
            "taobao_user_id": "1",
            "taobao_user_nick": "foobar",
            "w1_expires_in": 1800,
            "re_expires_in": 0,
            "r2_expires_in": 0,
            "expires_in": 86400,
            "r1_expires_in": 1800,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .oauth import BaseOAuth1, BaseOAuth2


class VimeoOAuth1(BaseOAuth1):
    """Vimeo OAuth authentication backend"""

    name = "vimeo"
    AUTHORIZATION_URL = "https://vimeo.com/oauth/authorize"
    REQUEST_TOKEN_URL = "https://vimeo.com/oauth/request_token"
    ACCESS_TOKEN_URL = "https://vimeo.com/oauth/access_token"

    def get_user_id(self, details, response):
        return response.get("person", {}).get("id")

    def get_user_details(self, response):
        """Return user details from Twitter account"""
        person = response.get("person", {})
        fullname, first_name, last_name = self.get_user_names(
            person.get("display_name", "")
        )
        return {
            "username": person.get("username", ""),
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "https://vimeo.com/api/rest/v2",
            params={"format": "json", "method": "vimeo.people.getInfo"},
            auth=self.oauth_auth(access_token),
        )


class VimeoOAuth2(BaseOAuth2):
    """Vimeo OAuth2 authentication backend"""

    name = "vimeo-oauth2"
    AUTHORIZATION_URL = "https://api.vimeo.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.vimeo.com/oauth/access_token"
    REFRESH_TOKEN_URL = "https://api.vimeo.com/oauth/request_token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    API_ACCEPT_HEADER = {"Accept": "application/vnd.vimeo.*+json;version=3.0"}

    def get_redirect_uri(self, state=None):
        """
        Build redirect with redirect_state parameter.

        @Vimeo API 3 requires exact redirect uri without additional
        additional state parameter included
        """
        return self.redirect_uri

    def get_user_id(self, details, response):
        """Return user id"""
        try:
            user_id = response.get("user", {})["uri"].split("/")[-1]
        except KeyError:
            user_id = None
        return user_id

    def get_user_details(self, response):
        """Return user details from account"""
        user = response.get("user", {})
        fullname, first_name, last_name = self.get_user_names(user.get("name", ""))
        return {
            "username": fullname,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "https://api.vimeo.com/me",
            params={"access_token": access_token},
            headers=VimeoOAuth2.API_ACCEPT_HEADER,
        )



from .oauth import BaseOAuth2


class MineIDOAuth2(BaseOAuth2):
    """MineID OAuth2 authentication backend"""

    name = "mineid"
    _AUTHORIZATION_URL = "%(scheme)s://%(host)s/oauth/authorize"
    _ACCESS_TOKEN_URL = "%(scheme)s://%(host)s/oauth/access_token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = []

    def get_user_details(self, response):
        """Return user details"""
        return {"email": response.get("email"), "username": response.get("email")}

    def user_data(self, access_token, *args, **kwargs):
        return self._user_data(access_token)

    def _user_data(self, access_token, path=None):
        url = "{scheme}://{host}/api/user".format(**self.get_mineid_url_params())
        return self.get_json(url, params={"access_token": access_token})

    @property
    def AUTHORIZATION_URL(self):
        return self._AUTHORIZATION_URL % self.get_mineid_url_params()

    @property
    def ACCESS_TOKEN_URL(self):
        return self._ACCESS_TOKEN_URL % self.get_mineid_url_params()

    def get_mineid_url_params(self):
        return {
            "host": self.setting("HOST", "www.mineid.org"),
            "scheme": self.setting("SCHEME", "https"),
        }



"""
Appsfueld OAuth2 backend (with sandbox mode support), docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/appsfuel.html
"""

from .oauth import BaseOAuth2


class AppsfuelOAuth2(BaseOAuth2):
    name = "appsfuel"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "http://app.appsfuel.com/content/permission"
    ACCESS_TOKEN_URL = "https://api.appsfuel.com/v1/live/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    USER_DETAILS_URL = "https://api.appsfuel.com/v1/live/user"

    def get_user_details(self, response):
        """Return user details from Appsfuel account"""
        email = response.get("email", "")
        username = email.split("@")[0] if email else ""
        fullname, first_name, last_name = self.get_user_names(
            response.get("display_name", "")
        )
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            self.USER_DETAILS_URL, params={"access_token": access_token}
        )


class AppsfuelOAuth2Sandbox(AppsfuelOAuth2):
    name = "appsfuel-sandbox"
    AUTHORIZATION_URL = "https://api.appsfuel.com/v1/sandbox/choose"
    ACCESS_TOKEN_URL = "https://api.appsfuel.com/v1/sandbox/oauth/token"
    USER_DETAILS_URL = "https://api.appsfuel.com/v1/sandbox/user"



from .oauth import BaseOAuth2


class GoClioOAuth2(BaseOAuth2):
    name = "goclio"
    AUTHORIZATION_URL = "https://app.goclio.com/oauth/authorize/"
    ACCESS_TOKEN_METHOD = "POST"
    ACCESS_TOKEN_URL = "https://app.goclio.com/oauth/token/"
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def get_user_details(self, response):
        """Return user details from GoClio account"""
        user = response.get("user", {})
        username = user.get("id", None)
        email = user.get("email", None)
        first_name, last_name = (
            user.get("first_name", None),
            user.get("last_name", None),
        )
        fullname = f"{first_name} {last_name}"

        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://app.goclio.com/api/v2/users/who_am_i",
            params={"access_token": access_token},
        )

    def get_user_id(self, details, response):
        return response.get("user", {}).get("id")



import json
import unittest
from urllib.parse import urlparse

import requests
from httpretty import HTTPretty

from ...actions import do_auth, do_complete
from ...utils import module_member, parse_qs
from ..models import TestAssociation, TestNonce, TestStorage, TestUserSocialAuth, User
from ..strategy import TestStrategy


class BaseActionTest(unittest.TestCase):
    user_data_url = "https://api.github.com/user"
    login_redirect_url = "/success"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def __init__(self, *args, **kwargs):
        self.strategy = None
        self.backend = None
        super().__init__(*args, **kwargs)

    def setUp(self):
        HTTPretty.enable()
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        Backend = module_member("social_core.backends.github.GithubOAuth2")
        self.strategy = self.strategy or TestStrategy(TestStorage)
        self.backend = self.backend or Backend(
            self.strategy, redirect_uri="/complete/github"
        )
        self.user = None

    def tearDown(self):
        self.backend = None
        self.strategy = None
        self.user = None
        User.reset_cache()
        User.set_active(True)
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        HTTPretty.disable()

    def do_login(
        self, after_complete_checks=True, user_data_body=None, expected_username=None
    ):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_KEY": "a-key",
                "SOCIAL_AUTH_GITHUB_SECRET": "a-secret-key",
                "SOCIAL_AUTH_LOGIN_REDIRECT_URL": self.login_redirect_url,
                "SOCIAL_AUTH_AUTHENTICATION_BACKENDS": (
                    "social_core.backends.github.GithubOAuth2",
                ),
            }
        )
        start_url = do_auth(self.backend).url
        target_url = self.strategy.build_absolute_uri("/complete/github/?code=foobar")

        start_query = parse_qs(urlparse(start_url).query)
        location_url = (
            target_url
            + ("&" if "?" in target_url else "?")
            + "state="
            + start_query["state"]
        )
        location_query = parse_qs(urlparse(location_url).query)

        HTTPretty.register_uri(
            HTTPretty.GET, start_url, status=301, location=location_url
        )
        HTTPretty.register_uri(HTTPretty.GET, location_url, status=200, body="foobar")

        response = requests.get(start_url)
        self.assertEqual(response.url, location_url)
        self.assertEqual(response.text, "foobar")

        HTTPretty.register_uri(
            HTTPretty.POST,
            uri=self.backend.ACCESS_TOKEN_URL,
            status=200,
            body=self.access_token_body or "",
            content_type="text/json",
        )

        if self.user_data_url:
            user_data_body = user_data_body or self.user_data_body or ""
            HTTPretty.register_uri(
                HTTPretty.GET,
                self.user_data_url,
                body=user_data_body,
                content_type="text/json",
            )
        self.strategy.set_request_data(location_query, self.backend)

        def _login(backend, user, social_user):
            backend.strategy.session_set("username", user.username)
            user_email = getattr(user, "email", None)
            if user_email:
                backend.strategy.session_set("email", user_email)

        redirect = do_complete(self.backend, user=self.user, login=_login)

        if after_complete_checks:
            self.assertEqual(
                self.strategy.session_get("username"),
                expected_username or self.expected_username,
            )
            self.assertEqual(redirect.url, self.login_redirect_url)
        return redirect

    def do_login_with_partial_pipeline(self, before_complete=None):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GITHUB_KEY": "a-key",
                "SOCIAL_AUTH_GITHUB_SECRET": "a-secret-key",
                "SOCIAL_AUTH_LOGIN_REDIRECT_URL": self.login_redirect_url,
                "SOCIAL_AUTH_AUTHENTICATION_BACKENDS": (
                    "social_core.backends.github.GithubOAuth2",
                ),
                "SOCIAL_AUTH_PIPELINE": (
                    "social_core.pipeline.social_auth.social_details",
                    "social_core.pipeline.social_auth.social_uid",
                    "social_core.pipeline.social_auth.auth_allowed",
                    "social_core.tests.pipeline.ask_for_password",
                    "social_core.pipeline.social_auth.social_user",
                    "social_core.pipeline.user.get_username",
                    "social_core.pipeline.user.create_user",
                    "social_core.pipeline.social_auth.associate_user",
                    "social_core.pipeline.social_auth.load_extra_data",
                    "social_core.tests.pipeline.set_password",
                    "social_core.pipeline.user.user_details",
                ),
            }
        )
        start_url = do_auth(self.backend).url
        target_url = self.strategy.build_absolute_uri("/complete/github/?code=foobar")

        start_query = parse_qs(urlparse(start_url).query)
        location_url = (
            target_url
            + ("&" if "?" in target_url else "?")
            + "state="
            + start_query["state"]
        )
        location_query = parse_qs(urlparse(location_url).query)

        HTTPretty.register_uri(
            HTTPretty.GET, start_url, status=301, location=location_url
        )
        HTTPretty.register_uri(HTTPretty.GET, location_url, status=200, body="foobar")

        response = requests.get(start_url)
        self.assertEqual(response.url, location_url)
        self.assertEqual(response.text, "foobar")

        HTTPretty.register_uri(
            HTTPretty.GET,
            uri=self.backend.ACCESS_TOKEN_URL,
            status=200,
            body=self.access_token_body or "",
            content_type="text/json",
        )

        if self.user_data_url:
            HTTPretty.register_uri(
                HTTPretty.GET,
                self.user_data_url,
                body=self.user_data_body or "",
                content_type="text/json",
            )
        self.strategy.set_request_data(location_query, self.backend)

        def _login(backend, user, social_user):
            backend.strategy.session_set("username", user.username)
            user_email = getattr(user, "email", None)
            if user_email:
                backend.strategy.session_set("email", user_email)

        redirect = do_complete(self.backend, user=self.user, login=_login)
        url = self.strategy.build_absolute_uri("/password")
        self.assertEqual(redirect.url, url)
        HTTPretty.register_uri(HTTPretty.GET, redirect.url, status=200, body="foobar")
        HTTPretty.register_uri(HTTPretty.POST, redirect.url, status=200)

        password = "foobar"
        requests.get(url)
        requests.post(url, data={"password": password})
        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data["password"], password)
        self.strategy.session_set("password", data["password"])

        if before_complete:
            before_complete()
        redirect = do_complete(self.backend, user=self.user, login=_login)
        self.assertEqual(self.strategy.session_get("username"), self.expected_username)
        self.assertEqual(redirect.url, self.login_redirect_url)

    def _logout(self, backend):
        backend.strategy.session_set("username", None)



"""
OpenStreetMap OAuth support.

This adds support for OpenStreetMap OAuth service. An application must be
registered first on OpenStreetMap and the settings
SOCIAL_AUTH_OPENSTREETMAP_KEY and SOCIAL_AUTH_OPENSTREETMAP_SECRET
must be defined with the corresponding values.

More info: https://wiki.openstreetmap.org/wiki/OAuth
"""

from xml.dom import minidom

from .oauth import BaseOAuth1


class OpenStreetMapOAuth(BaseOAuth1):
    """OpenStreetMap OAuth authentication backend"""

    name = "openstreetmap"
    AUTHORIZATION_URL = "https://www.openstreetmap.org/oauth/authorize"
    REQUEST_TOKEN_URL = "https://www.openstreetmap.org/oauth/request_token"
    ACCESS_TOKEN_URL = "https://www.openstreetmap.org/oauth/access_token"
    EXTRA_DATA = [
        ("id", "id"),
        ("avatar", "avatar"),
        ("account_created", "account_created"),
    ]

    def get_user_details(self, response):
        """Return user details from OpenStreetMap account"""
        return {
            "username": response["username"],
            "email": "",
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        response = self.oauth_request(
            access_token, "https://api.openstreetmap.org/api/0.6/user/details"
        )
        try:
            dom = minidom.parseString(response.content)
        except ValueError:
            return None
        user = dom.getElementsByTagName("user")[0]
        try:
            avatar = dom.getElementsByTagName("img")[0].getAttribute("href")
        except IndexError:
            avatar = None
        return {
            "id": user.getAttribute("id"),
            "username": user.getAttribute("display_name"),
            "account_created": user.getAttribute("account_created"),
            "avatar": avatar,
        }



"""
Trello OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/trello.html
"""

from .oauth import BaseOAuth1


class TrelloOAuth(BaseOAuth1):
    """Trello OAuth authentication backend"""

    name = "trello"
    ID_KEY = "username"
    AUTHORIZATION_URL = "https://trello.com/1/OAuthAuthorizeToken"
    REQUEST_TOKEN_URL = "https://trello.com/1/OAuthGetRequestToken"
    ACCESS_TOKEN_URL = "https://trello.com/1/OAuthGetAccessToken"

    EXTRA_DATA = [
        ("username", "username"),
        ("email", "email"),
        ("fullName", "fullName"),
    ]

    def get_user_details(self, response):
        """Return user details from Trello account"""
        fullname, first_name, last_name = self.get_user_names(response.get("fullName"))
        return {
            "username": response.get("username"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token):
        """Return user data provided"""
        url = "https://trello.com/1/members/me"
        try:
            return self.get_json(url, auth=self.oauth_auth(access_token))
        except ValueError:
            return None

    def auth_extra_arguments(self):
        return {
            "name": self.setting("APP_NAME", ""),
            # trello default expiration is '30days'
            "expiration": self.setting("EXPIRATION", "never"),
        }



"""
Quizlet OAuth2 Sign-in backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/quizlet.html
"""

from .oauth import BaseOAuth2


class QuizletOAuth2(BaseOAuth2):
    """Quizlet OAuth2"""

    name = "quizlet"
    ID_KEY = "user_id"
    API_URL = "https://api.quizlet.com/2.0/"
    AUTHORIZATION_URL = "https://quizlet.com/authorize"
    ACCESS_TOKEN_URL = "https://api.quizlet.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = " "
    DEFAULT_SCOPE = ["read"]

    def get_user_details(self, response):
        """Return user details from Quizlet account"""
        return {"username": response.get("user_id")}



# pyright: reportAttributeAccessIssue=false

import json

import requests
from httpretty import HTTPretty

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class StripeOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.stripe.StripeOAuth2"
    account_data_url = "https://api.stripe.com/v1/account"
    access_token_body = json.dumps(
        {
            "stripe_publishable_key": "pk_test_foobar",
            "access_token": "foobar",
            "livemode": False,
            "token_type": "bearer",
            "scope": "read_only",
            "refresh_token": "rt_foobar",
            "stripe_user_id": "acct_foobar",
        }
    )
    expected_username = "acct_foobar"
    user_data_body = json.dumps(
        {
            "id": "acct_1LUYJiECsxMRIeT8",
            "object": "account",
            "country": "FR",
            "created": 1659974194,
            "default_currency": "eur",
            "details_submitted": True,
            "email": "foobar@yahoo.com",
            "type": "express",
        }
    )

    def setUp(self):
        super().setUp()
        HTTPretty.register_uri(
            HTTPretty.GET, self.account_data_url, status=200, body=self.user_data_body
        )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_get_user_details(self):
        response = requests.get(self.account_data_url)
        user_details = self.backend.get_user_details(response.json())
        self.assertEqual(user_details["email"], "foobar@yahoo.com")



import requests

from ..exceptions import AuthFailed
from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class UntappdOAuth2(BaseOAuth2):
    """Untappd OAuth2 authentication backend"""

    name = "untappd"
    AUTHORIZATION_URL = "https://untappd.com/oauth/authenticate/"
    ACCESS_TOKEN_URL = "https://untappd.com/oauth/authorize/"
    BASE_API_URL = "https://api.untappd.com"
    USER_INFO_URL = BASE_API_URL + "/v4/user/info/"
    ACCESS_TOKEN_METHOD = "GET"
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ("id", "id"),
        ("bio", "bio"),
        ("date_joined", "date_joined"),
        ("location", "location"),
        ("url", "url"),
        ("user_avatar", "user_avatar"),
        ("user_avatar_hd", "user_avatar_hd"),
        ("user_cover_photo", "user_cover_photo"),
    ]

    def auth_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "client_id": client_id,
            "redirect_url": self.get_redirect_uri(),
            "response_type": self.RESPONSE_TYPE,
        }

    def process_error(self, data):
        """
        All errors from Untappd are contained in the 'meta' key of the
        response.
        """
        response_code = data.get("meta", {}).get("http_code")
        if response_code is not None and response_code != requests.codes.ok:
            raise AuthFailed(self, data["meta"]["error_detail"])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        client_id, client_secret = self.get_key_and_secret()
        code = self.data.get("code")

        self.process_error(self.data)

        # Untapped sends the access token request with URL parameters,
        # not a body
        response = self.request_access_token(
            self.access_token_url(),
            method=self.ACCESS_TOKEN_METHOD,
            params={
                "response_type": "code",
                "code": code,
                "client_id": client_id,
                "client_secret": client_secret,
                "redirect_url": self.get_redirect_uri(),
            },
        )

        self.process_error(response)

        # Both the access_token and the rest of the response are
        # buried in the 'response' key
        return self.do_auth(
            response["response"]["access_token"],
            response=response["response"],
            *args,
            **kwargs,
        )

    def get_user_details(self, response):
        """Return user details from an Untappd account"""
        # Start with the user data as it was returned
        user_data = response["user"]

        # Make a few updates to match expected key names
        user_data.update(
            {
                "username": user_data.get("user_name"),
                "email": user_data.get("settings", {}).get("email_address", ""),
                "first_name": user_data.get("first_name"),
                "last_name": user_data.get("last_name"),
                "fullname": " ".join(
                    [user_data.get("first_name"), user_data.get("last_name")]
                ),
            }
        )
        return user_data

    def get_user_id(self, details, response):
        """
        Return a unique ID for the current user, by default from
        server response.
        """
        return response["user"].get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = self.get_json(
            self.USER_INFO_URL, params={"access_token": access_token, "compact": "true"}
        )
        self.process_error(response)

        # The response data is buried in the 'response' key
        return response["response"]



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class WLCGOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.wlcg.WLCGOAuth2"
    user_data_url = "https://wlcg.cloud.cnaf.infn.it/userinfo"
    expected_username = "foo@bar.com"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
        }
    )
    user_data_body = json.dumps(
        {
            "email": "foo@bar.com",
            "family_name": "Bar",
            "given_name": "Foo",
            "name": "Foo Bar",
            "email_verified": True,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Discord Auth OAuth2 backend, docs at:
    https://discord.com/developers/docs/topics/oauth2
"""

from .oauth import BaseOAuth2


class DiscordOAuth2(BaseOAuth2):
    name = "discord"
    HOSTNAME = "discord.com"
    AUTHORIZATION_URL = f"https://{HOSTNAME}/api/oauth2/authorize"
    ACCESS_TOKEN_URL = f"https://{HOSTNAME}/api/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    REVOKE_TOKEN_URL = f"https://{HOSTNAME}/api/oauth2/token/revoke"
    REVOKE_TOKEN_METHOD = "GET"
    DEFAULT_SCOPE = ["identify"]
    SCOPE_SEPARATOR = "+"
    REDIRECT_STATE = False
    EXTRA_DATA = [("expires_in", "expires"), ("refresh_token", "refresh_token")]

    def get_user_details(self, response):
        return {
            "username": response.get("username"),
            "email": response.get("email") or "",
        }

    def user_data(self, access_token, *args, **kwargs):
        url = f"https://{self.HOSTNAME}/api/users/@me"
        auth_header = {"Authorization": f"Bearer {access_token}"}
        return self.get_json(url, headers=auth_header)



import requests
from httpretty import HTTPretty

from ...actions import do_disconnect
from ...exceptions import NotAllowedToDisconnect
from ...utils import parse_qs
from ..models import TestUserSocialAuth, User
from .actions import BaseActionTest


class DisconnectActionTest(BaseActionTest):
    def test_not_allowed_to_disconnect(self):
        self.do_login()
        user = User.get(self.expected_username)
        with self.assertRaisesRegex(
            NotAllowedToDisconnect, "This account is not allowed to be disconnected."
        ):
            do_disconnect(self.backend, user)

    def test_disconnect(self):
        self.do_login()
        user = User.get(self.expected_username)
        user.password = "password"
        do_disconnect(self.backend, user)
        self.assertEqual(len(user.social), 0)

    def test_disconnect_with_association_id(self):
        self.do_login()
        user = User.get(self.expected_username)
        user.password = "password"
        association_id = user.social[0].id
        second_usa = TestUserSocialAuth(user, user.social[0].provider, "uid2")
        self.assertEqual(len(user.social), 2)
        do_disconnect(self.backend, user, association_id)
        self.assertEqual(len(user.social), 1)
        self.assertEqual(user.social[0], second_usa)

    def test_disconnect_with_partial_pipeline(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_DISCONNECT_PIPELINE": (
                    "social_core.tests.pipeline.ask_for_password",
                    "social_core.tests.pipeline.set_password",
                    "social_core.pipeline.disconnect.allowed_to_disconnect",
                    "social_core.pipeline.disconnect.get_entries",
                    "social_core.pipeline.disconnect.revoke_tokens",
                    "social_core.pipeline.disconnect.disconnect",
                )
            }
        )
        self.do_login()
        user = User.get(self.expected_username)
        redirect = do_disconnect(self.backend, user)

        url = self.strategy.build_absolute_uri("/password")
        self.assertEqual(redirect.url, url)
        HTTPretty.register_uri(HTTPretty.GET, redirect.url, status=200, body="foobar")
        HTTPretty.register_uri(HTTPretty.POST, redirect.url, status=200)

        password = "foobar"
        requests.get(url)
        requests.post(url, data={"password": password})
        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data["password"], password)
        self.strategy.session_set("password", data["password"])

        redirect = do_disconnect(self.backend, user)
        self.assertEqual(len(user.social), 0)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class CourseraOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.coursera.CourseraOAuth2"
    user_data_url = "https://api.coursera.org/api/externalBasicProfiles.v1?q=me"
    expected_username = "560e7ed2076e0d589e88bd74b6aad4b7"
    access_token_body = json.dumps(
        {"access_token": "foobar", "token_type": "Bearer", "expires_in": 1795}
    )
    request_token_body = json.dumps(
        {
            "code": "foobar-code",
            "client_id": "foobar-client-id",
            "client_secret": "foobar-client-secret",
            "redirect_uri": "http://localhost:8000/accounts/coursera/",
            "grant_type": "authorization_code",
        }
    )
    user_data_body = json.dumps(
        {
            "token_type": "Bearer",
            "paging": None,
            "elements": [{"id": "560e7ed2076e0d589e88bd74b6aad4b7"}],
            "access_token": "foobar",
            "expires_in": 1800,
            "linked": None,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Behance OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/behance.html
"""

from .oauth import BaseOAuth2


class BehanceOAuth2(BaseOAuth2):
    """Behance OAuth authentication backend"""

    name = "behance"
    AUTHORIZATION_URL = "https://www.behance.net/v2/oauth/authenticate"
    ACCESS_TOKEN_URL = "https://www.behance.net/v2/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = "|"
    EXTRA_DATA = [("username", "username")]
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response["user"]["id"]

    def get_user_details(self, response):
        """Return user details from Behance account"""
        user = response["user"]
        fullname, first_name, last_name = self.get_user_names(
            user["display_name"], user["first_name"], user["last_name"]
        )
        return {
            "username": user["username"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": "",
        }

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        # Pull up the embedded user attributes so they can be found as extra
        # data. See the example token response for possible attributes:
        # http://www.behance.net/dev/authentication#step-by-step
        data = response.copy()
        data.update(response["user"])
        return super().extra_data(user, uid, data, details, *args, **kwargs)



# pyright: reportAttributeAccessIssue=false

import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class CognitoAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.cognito.CognitoOAuth2"
    pool_domain = "https://social_core.auth.eu-west-1.amazoncognito.com"
    expected_username = "cognito.account.ABCDE1234"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "given_name": "John",
            "family_name": "Doe",
            "username": "cognito.account.ABCDE1234",
            "email": "john@doe.test",
        }
    )

    @property
    def user_data_url(self):
        return self.backend.user_data_url()

    def extra_settings(self):
        settings = super().extra_settings()
        assert self.name, "subclasses must set the name attribute"
        settings.update(
            {
                "SOCIAL_AUTH_" + self.name + "_POOL_DOMAIN": self.pool_domain,
            }
        )
        return settings

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Twitter OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/twitter.html
"""

from ..exceptions import AuthCanceled
from .oauth import BaseOAuth1


class TwitterOAuth(BaseOAuth1):
    """Twitter OAuth authentication backend"""

    name = "twitter"
    EXTRA_DATA = [("id", "id")]
    REQUEST_TOKEN_METHOD = "POST"
    ACCESS_TOKEN_METHOD = "POST"
    AUTHORIZATION_URL = "https://api.twitter.com/oauth/authenticate"
    REQUEST_TOKEN_URL = "https://api.twitter.com/oauth/request_token"
    ACCESS_TOKEN_URL = "https://api.twitter.com/oauth/access_token"
    REDIRECT_STATE = True

    def process_error(self, data):
        if "denied" in data:
            raise AuthCanceled(self)
        super().process_error(data)

    def get_user_details(self, response):
        """Return user details from Twitter account"""
        fullname, first_name, last_name = self.get_user_names(response["name"])
        return {
            "username": response["screen_name"],
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "https://api.twitter.com/1.1/account/verify_credentials.json",
            params={"include_email": "true"},
            auth=self.oauth_auth(access_token),
        )



import unittest
from unittest.mock import Mock

from ..utils import (
    build_absolute_uri,
    partial_pipeline_data,
    sanitize_redirect,
    slugify,
    user_is_active,
    user_is_authenticated,
)
from .models import TestPartial


class SanitizeRedirectTest(unittest.TestCase):
    def test_none_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], None), None)

    def test_empty_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], ""), None)

    def test_dict_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], {}), None)

    def test_invalid_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], {"foo": "bar"}), None)

    def test_wrong_path_redirect(self):
        self.assertEqual(
            sanitize_redirect(["myapp.com"], "http://notmyapp.com/path/"), None
        )

    def test_invalid_evil_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], "///evil.com"), None)

    def test_valid_absolute_redirect(self):
        self.assertEqual(
            sanitize_redirect(["myapp.com"], "http://myapp.com/path/"),
            "http://myapp.com/path/",
        )

    def test_valid_relative_redirect(self):
        self.assertEqual(sanitize_redirect(["myapp.com"], "/path/"), "/path/")

    def test_multiple_hosts(self):
        allowed_hosts = ["myapp1.com", "myapp2.com"]
        for host in allowed_hosts:
            url = f"http://{host}/path/"
            self.assertEqual(sanitize_redirect(allowed_hosts, url), url)

    def test_multiple_hosts_wrong_host(self):
        self.assertEqual(
            sanitize_redirect(
                ["myapp1.com", "myapp2.com"], "http://notmyapp.com/path/"
            ),
            None,
        )


class UserIsAuthenticatedTest(unittest.TestCase):
    def test_user_is_none(self):
        self.assertEqual(user_is_authenticated(None), False)

    def test_user_is_not_none(self):
        self.assertEqual(user_is_authenticated(object()), True)

    def test_user_has_is_authenticated(self):
        class User:
            is_authenticated = True

        self.assertEqual(user_is_authenticated(User()), True)

    def test_user_has_is_authenticated_callable(self):
        class User:
            def is_authenticated(self):
                return True

        self.assertEqual(user_is_authenticated(User()), True)


class UserIsActiveTest(unittest.TestCase):
    def test_user_is_none(self):
        self.assertEqual(user_is_active(None), False)

    def test_user_is_not_none(self):
        self.assertEqual(user_is_active(object()), True)

    def test_user_has_is_active(self):
        class User:
            is_active = True

        self.assertEqual(user_is_active(User()), True)

    def test_user_has_is_active_callable(self):
        class User:
            def is_active(self):
                return True

        self.assertEqual(user_is_active(User()), True)


class SlugifyTest(unittest.TestCase):
    def test_slugify_formats(self):
        self.assertEqual(slugify("FooBar"), "foobar")
        self.assertEqual(slugify("Foo Bar"), "foo-bar")
        self.assertEqual(slugify("Foo (Bar)"), "foo-bar")


class BuildAbsoluteURITest(unittest.TestCase):
    def setUp(self):
        self.host = "http://foobar.com"

    def tearDown(self):
        self.host = None

    def test_path_none(self):
        self.assertEqual(build_absolute_uri(self.host), self.host)

    def test_path_empty(self):
        self.assertEqual(build_absolute_uri(self.host, ""), self.host)

    def test_path_http(self):
        self.assertEqual(
            build_absolute_uri(self.host, "http://barfoo.com"), "http://barfoo.com"
        )

    def test_path_https(self):
        self.assertEqual(
            build_absolute_uri(self.host, "https://barfoo.com"), "https://barfoo.com"
        )

    def test_host_ends_with_slash_and_path_starts_with_slash(self):
        assert self.host, "Subclasses must set the host attribute"
        self.assertEqual(
            build_absolute_uri(self.host + "/", "/foo/bar"), "http://foobar.com/foo/bar"
        )

    def test_absolute_uri(self):
        self.assertEqual(
            build_absolute_uri(self.host, "/foo/bar"), "http://foobar.com/foo/bar"
        )


class PartialPipelineData(unittest.TestCase):
    def test_returns_partial_when_uid_and_email_do_match(self):
        email = "foo@example.com"
        backend = self._backend({"uid": email})
        backend.strategy.request_data.return_value = {backend.ID_KEY: email}
        key, val = ("foo", "bar")
        partial = partial_pipeline_data(backend, None, *(), **{key: val})
        self.assertTrue(key in partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_clean_pipeline_when_uid_does_not_match(self):
        backend = self._backend({"uid": "foo@example.com"})
        backend.strategy.request_data.return_value = {backend.ID_KEY: "bar@example.com"}
        key, val = ("foo", "bar")
        partial = partial_pipeline_data(backend, None, *(), **{key: val})
        self.assertIsNone(partial)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 1)

    def test_kwargs_included_in_result(self):
        backend = self._backend()
        key, val = ("foo", "bar")
        partial = partial_pipeline_data(backend, None, *(), **{key: val})
        self.assertTrue(key in partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_update_user(self):
        user = object()
        backend = self._backend(session_kwargs={"user": None})
        partial = partial_pipeline_data(backend, user)
        self.assertTrue("user" in partial.kwargs)
        self.assertEqual(partial.kwargs["user"], user)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def _backend(self, session_kwargs=None):
        backend = Mock()
        backend.ID_KEY = "email"
        backend.name = "mock-backend"

        strategy = Mock()
        strategy.request = None
        strategy.request_data.return_value = {}
        strategy.session_get.return_value = object()
        strategy.partial_load.return_value = TestPartial.prepare(
            backend.name, 0, {"args": [], "kwargs": session_kwargs or {}}
        )

        backend.strategy = strategy
        return backend



import datetime
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...exceptions import AuthMissingParameter
from .open_id import OpenIdTest

JANRAIN_NONCE = datetime.datetime.now(datetime.timezone.utc).strftime(
    "%Y-%m-%dT%H:%M:%SZ"
)


class LiveJournalOpenIdTest(OpenIdTest):
    backend_path = "social_core.backends.livejournal.LiveJournalOpenId"
    expected_username = "foobar"
    discovery_body = """<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">
    <XRD>
        <Service priority="0">
            <Type>http://specs.openid.net/auth/2.0/signon</Type>
            <URI>http://www.livejournal.com/openid/server.bml</URI>
            <LocalID>http://foobar.livejournal.com/</LocalID>
        </Service>
    </XRD>
</xrds:XRDS>"""
    server_response = urlencode(
        {
            "janrain_nonce": JANRAIN_NONCE,
            "openid.mode": "id_res",
            "openid.claimed_id": "http://foobar.livejournal.com/",
            "openid.identity": "http://foobar.livejournal.com/",
            "openid.op_endpoint": "http://www.livejournal.com/openid/server.bml",
            "openid.return_to": "http://myapp.com/complete/livejournal/?"
            "janrain_nonce=" + JANRAIN_NONCE,
            "openid.response_nonce": JANRAIN_NONCE + "wGp2rj",
            "openid.assoc_handle": "1364932966:ZTiur8sem3r2jzZougMZ:4d1cc3b44e",
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.signed": "mode,claimed_id,identity,op_endpoint,return_to,"
            "response_nonce,assoc_handle",
            "openid.sig": "Z8MOozVPTOBhHG5ZS1NeGofxs1Q=",
        }
    )
    server_bml_body = """assoc_handle:1364935340:ZhruPQ7DJ9eGgUkeUA9A:27f8c32464
assoc_type:HMAC-SHA1
dh_server_public:WzsRyLomvAV3vwvGUrfzXDgfqnTF+m1l3JWb55fyHO7visPT4tmQiTjqFFnSVAtAOvQzoViMiZQisxNwnqSK4lYexoez1z6pP5ry3pqxJAEYj60vFGvRztictEo0brjhmO1SNfjK1ppjOymdykqLpZeaL5fsuLtMCwTnR/JQZVA=
enc_mac_key:LiOEVlLJSVUqfNvb5zPd76nEfvc=
expires_in:1207060
ns:http://specs.openid.net/auth/2.0
session_type:DH-SHA1
"""

    def openid_url(self):
        return super().openid_url() + "/data/yadis"

    def post_start(self):
        self.strategy.remove_from_request_data("openid_lj_user")

    def _setup_handlers(self):
        HTTPretty.register_uri(
            HTTPretty.POST,
            "http://www.livejournal.com/openid/server.bml",
            headers={
                "Accept-Encoding": "identity",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            status=200,
            body=self.server_bml_body,
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            "http://foobar.livejournal.com/",
            headers={
                "Accept-Encoding": "identity",
                "Accept": "text/html; q=0.3,"
                "application/xhtml+xml; q=0.5,"
                "application/xrds+xml",
            },
            status=200,
            body=self.discovery_body,
        )

    def test_login(self):
        self.strategy.set_request_data({"openid_lj_user": "foobar"}, self.backend)
        self._setup_handlers()
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_request_data({"openid_lj_user": "foobar"}, self.backend)
        self._setup_handlers()
        self.do_partial_pipeline()

    def test_failed_login(self):
        self._setup_handlers()
        with self.assertRaises(AuthMissingParameter):
            self.do_login()



"""
Stackoverflow OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/stackoverflow.html
"""

from .oauth import BaseOAuth2


class StackoverflowOAuth2(BaseOAuth2):
    """Stackoverflow OAuth2 authentication backend"""

    name = "stackoverflow"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://stackexchange.com/oauth"
    ACCESS_TOKEN_URL = "https://stackexchange.com/oauth/access_token"
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = [("id", "id"), ("expires", "expires")]

    def get_user_details(self, response):
        """Return user details from Stackoverflow account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("display_name")
        )
        return {
            "username": response.get("link").rsplit("/", 1)[-1],
            "full_name": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.stackexchange.com/2.1/me",
            params={
                "site": "stackoverflow",
                "access_token": access_token,
                "key": self.setting("API_KEY"),
            },
        )["items"][0]

    def request_access_token(self, *args, **kwargs):
        return self.get_querystring(*args, **kwargs)



"""
Uber OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/uber.html
"""

from .oauth import BaseOAuth2


class UberOAuth2(BaseOAuth2):
    name = "uber"
    ID_KEY = "uuid"
    SCOPE_SEPARATOR = " "
    AUTHORIZATION_URL = "https://login.uber.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://login.uber.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"

    def auth_complete_credentials(self):
        return self.get_key_and_secret()

    def get_user_details(self, response):
        """Return user details from Uber account"""
        email = response.get("email", "")
        fullname, first_name, last_name = self.get_user_names(
            "", response.get("first_name", ""), response.get("last_name", "")
        )
        return {
            "username": email,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = kwargs.pop("response")
        return self.get_json(
            "https://api.uber.com/v1/me",
            headers={
                "Authorization": "{} {}".format(
                    response.get("token_type"), access_token
                )
            },
        )



"""
Slack OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/slack.html
    https://api.slack.com/docs/oauth
"""

from .oauth import BaseOAuth2


class SlackOAuth2(BaseOAuth2):
    """Slack OAuth authentication backend"""

    name = "slack"
    AUTHORIZATION_URL = "https://slack.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://slack.com/api/oauth.access"
    ACCESS_TOKEN_METHOD = "POST"
    DEFAULT_SCOPE = ["identity.basic", "identity.email"]
    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False
    EXTRA_DATA = [("id", "id"), ("name", "name"), ("real_name", "real_name")]

    def auth_extra_arguments(self):
        params = super().auth_extra_arguments() or {}
        if self.setting("TEAM"):
            params["team"] = self.setting("TEAM")
        return params

    def get_user_details(self, response):
        """Return user details from Slack account"""
        # Build the username with the team $username@$team_url
        # Necessary to get unique names for all of slack
        user = response["user"]
        team = response.get("team")
        name = user["name"]
        email = user.get("email")
        username = (email and email.split("@", 1)[0]) or name
        fullname, first_name, last_name = self.get_user_names(name)

        if self.setting("USERNAME_WITH_TEAM", True) and team and "name" in team:
            username = "{}@{}".format(username, response["team"]["name"])

        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = self.get_json(
            "https://slack.com/api/users.identity",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        if not response.get("id", None):
            response["id"] = response["user"]["id"]
        return response



"""
Professionali OAuth 2.0 support.

This contribution adds support for professionali.ru OAuth 2.0.
Username is retrieved from the identity returned by server.
"""

from time import time

from ..utils import parse_qs
from .oauth import BaseOAuth2


class ProfessionaliOAuth2(BaseOAuth2):
    name = "professionali"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://api.professionali.ru/oauth/authorize.html"
    ACCESS_TOKEN_URL = "https://api.professionali.ru/oauth/getToken.json"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [("avatar_big", "avatar_big"), ("link", "link")]

    def get_user_details(self, response):
        first_name, last_name = map(response.get, ("firstname", "lastname"))
        email = ""
        if self.setting("FAKE_EMAIL"):
            email = f"{time()}@professionali.ru"
        return {
            "username": f"{last_name}_{first_name}",
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, response, *args, **kwargs):
        url = "https://api.professionali.ru/v6/users/get.json"
        fields = list(
            {
                "firstname",
                "lastname",
                "avatar_big",
                "link",
                *self.setting("EXTRA_DATA", []),
            }
        )
        params = {
            "fields": ",".join(fields),
            "access_token": access_token,
            "ids[]": response["user_id"],
        }
        try:
            return self.get_json(url, params)[0]
        except (TypeError, KeyError, OSError, ValueError, IndexError):
            return None

    def get_json(self, url, *args, **kwargs):
        return self.request(url, verify=False, *args, **kwargs).json()

    def get_querystring(self, url, *args, **kwargs):
        return parse_qs(self.request(url, verify=False, *args, **kwargs).text)
